#!/usr/bin/env pythonbutton_color
# -*- coding: utf-8 -*-

"""  

Higher level ui components / widgets like Edit_Panel, Diagram 

"""

import os

from copy               import copy
from typing             import override

from PyQt6.QtCore       import Qt, QEvent, QPoint, QRectF
from PyQt6.QtCore       import QSize, QMargins, QTimer, QPropertyAnimation, QAbstractAnimation
from PyQt6.QtWidgets    import QLayout, QGridLayout, QVBoxLayout, QHBoxLayout, QSizePolicy
from PyQt6.QtWidgets    import QMainWindow, QWidget, QDialog, QDialogButtonBox, QLabel, QMessageBox, QFrame
from PyQt6.QtWidgets    import QGraphicsOpacityEffect
from PyQt6.QtGui        import QGuiApplication, QScreen, QColor, QPalette, QPainterPath, QRegion, QTransform
from PyQt6              import sip

from base.widgets       import set_background, style
from base.widgets       import Widget, Label, CheckBox, size, Button, FieldI, SpaceR, Icon

import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.WARNING)


#------------------------------------------------------------------------------
# Utils for QMainWindow and QDialog  
#------------------------------------------------------------------------------

class Win_Util: 
    """ 
    Utility functions for window handling 
    """

    @staticmethod
    def set_initialWindowSize (qwindow : QWidget,
                               size : tuple | None = None,
                               size_frac : tuple | None = None,
                               pos : tuple | None = None,
                               pos_frac: tuple | None = None,
                               geometry : tuple | None = None,
                               maximize : bool = False):
        """
        Set size and position of Qt window in fraction of screensize or absolute
        """

        # geometry argument has priority 

        if geometry: 
            qwindow.setGeometry (*geometry)
            if maximize:
                qwindow.showMaximized()
            return
        else:  
            x, y, width, height = None, None, None, None
 
        # set size 

        if size_frac: 

            screen : QScreen = QGuiApplication.primaryScreen()
            screenGeometry = screen.geometry()
 
            width_frac, height_frac  = size_frac

            if width_frac:   width  = screenGeometry.width()  * width_frac
            if height_frac:  height = screenGeometry.height() * height_frac

        if size:
            width, height = size

        width  = int (width)  if width  is not None else 1000
        height = int (height) if height is not None else  700
        
        qwindow.resize (QSize(width, height))

        if maximize: 
            qwindow.showMaximized()

        # set position 

        if pos: 
            x, y = pos

        if pos_frac: 

            screen : QScreen = QGuiApplication.primaryScreen()
            screenGeometry = screen.geometry()
 
            x_frac = pos_frac[0]
            y_frac = pos_frac[1]
            if x_frac: x = screenGeometry.width()  * x_frac
            if y_frac: y = screenGeometry.height() * y_frac

        x = int (x) if x  is not None else 200
        y = int (y) if y is not None else  200
        
        qwindow.move (x, y)



#------------------------------------------------------------------------------
# Panels - QWidgets like a field group within a context 
#------------------------------------------------------------------------------


class Panel_Abstract (QWidget):
    """ 
    Superclass for other types of panels like Edit or Container
        - handle size of widget  
        - having title / header name 
        - has dataObject via getter (callable) 
    """

    name = "Panel"             # will be title 

    _width  = None
    _height = None 

    # ------------------------------------------------------

    @staticmethod
    def widgets_of_layout  (layout: QLayout) -> list[Widget]:
        """ list of Widgets defined in layout"""

        # iterate over layout - not self._panel.findChildren (Widget) as widgets are "deleted later"
        #       and can still exist although they are no more in the layout 

        widgets = []
        if isinstance (layout,QLayout): 
            for i in range(layout.count()):
                widget = layout.itemAt(i).widget()
                if isinstance (widget, Widget):                
                    widgets.append(widget)
                elif isinstance (widget,QWidget):                       # could be helper QWidget with seperate layout
                    widgets.extend (Panel_Abstract.widgets_of_layout (widget.layout()))
        return widgets

    # ------------------------------------------------------

    def __init__(self,  
                 app = None,
                 getter = None, 
                 width=None, 
                 hide=None,                                     # either callable or bool
                 height=None, 
                 title=None, **kwargs):
        # super().__init__(parent=app, **kwargs)
        super().__init__( **kwargs)

        self._app = app
        self._getter = getter

        if width is not None: 
            self._width = width
        if height is not None: 
            self._height = height

        # handle visibility  
        self._shouldBe_visible = True                               # defaulit visibilty of self 
        self._hidden_getter = None                                  # altern. getter to set visibility
        if isinstance(hide, bool):
            self._shouldBe_visible = not hide
        elif callable (hide):
            self._hidden_getter = hide

        # set width and height 
        Widget._set_width  (self, self._width)
        Widget._set_height (self, self._height)

        if title is not None: 
            self.name = title 


    def __repr__(self) -> str:
        # overwritten to get a nice print string 
        name = self.name if self.name else type(self).__name__
        return f"<Panel '{name}'>"


    @property
    def dataObject (self): 
        # to be overloaded - or implemented with semantic name 
        if callable(self._getter):
            return self._getter()
        else: 
            return self._getter

    @property 
    def shouldBe_visible (self) -> bool:
        """ True if self should be visible 
            - can be overridden to control visibility in subclass """
        
        if callable (self._hidden_getter):
            self._shouldBe_visible = not self._hidden_getter ()

        return self._shouldBe_visible


    def set_visibilty (self, aBool : bool):
        """ 
        set the visibility of self 
            - use this, when instances of Edit_Panel are used (not subclassing)
              to control hide/show
        """

        if self._shouldBe_visible != aBool:
            self._shouldBe_visible = aBool        
            self.setVisible (aBool)     


    @property 
    def _isDisabled (self) -> bool:
        """ True if the widgets of self are disabled  - can be overloaded """
        return False

    def set_background_color (self, darker_factor : int | None = None,
                                    color : QColor | int | None  = None,
                                    alpha : float | None = None):
        """ 
        Set background color of a QWidget either by
            - darker_factor > 100  
            - color: QColor or string for new color
            - alpha: transparency 0..1 
        """
        set_background (self, darker_factor=darker_factor, color=color, alpha=alpha)


#-------------------------------------------

class Container_Panel (Panel_Abstract):
    """ 
    Panel as container for other (Edit) panels  
    """

    def __init__(self, *args, 
                 layout : QLayout | None = None,
                 margins  : tuple = (0,0,0,0), 
                 spacing : int = 5,            
                 **kwargs):
        super().__init__(*args, **kwargs)

        if layout:
            layout.setContentsMargins (QMargins(*margins))  
            layout.setSpacing (spacing) 
            self.setLayout (layout)

        # initial visibility 

        if not self.shouldBe_visible:         
            self.setVisible (False)             # setVisible(True) results in a dummy window on startup 


    def refresh (self):
        """ refresh all child Panels self"""

        show_now = self.isHidden () and self.shouldBe_visible

        self.setVisible (self.shouldBe_visible)

        if self.shouldBe_visible:
            # first hide the now not visible panels so layout won't be stretched
            p : Edit_Panel
            for p in self.findChildren (Panel_Abstract):
                try:
                    if not p.shouldBe_visible: p.refresh() 
                except: 
                    logger.warning (f"{p} {hex(id(p))} (not visible) already deleted")

            # now show the now visible panels
            for p in self.findChildren (Panel_Abstract):
                if sip.isdeleted (p):
                    logger.warning (f"{p} {hex(id(p))} (visible) already deleted")
                else:
                    if p.shouldBe_visible: p.refresh (reinit_layout=show_now) 



class Edit_Panel (Panel_Abstract):
    """ 
    Panel with/out a title and an optional on/off switch 
    having a layout area for content  
    """

    _height = (150, None) 

    def __init__(self, *args, 
                 layout : QLayout | None = None,
                 lazy_layout = False, 
                 switchable : bool = False,
                 switched_on : bool = True, 
                 on_switched = None, 
                 hide_switched : bool = True,
                 has_head : bool = True,
                 main_margins  : tuple = (10, 5,10, 5),             
                 head_margins  : tuple = ( 0, 0, 0, 5),             
                 panel_margins : tuple = (15, 0, 0, 0),           
                 **kwargs):
        super().__init__(*args, **kwargs)

        self._switchable = switchable
        self._hide_switched = hide_switched
        self._switched_on = switched_on
        self._on_switched = on_switched 

        self._head  = None
        self._panel = None

        self._main_margins  = main_margins
        self._head_margins  = head_margins
        self._panel_margins = panel_margins

        # set background color depending light/dark mode

        if Widget.light_mode:
            self.set_background_color (darker_factor = 104)                 # make it darker 
        else: 
            self.set_background_color (darker_factor = 80)                  # make it lighter 

        # title layout - with optional on/off switch 

        if has_head and self.title_text() is not None: 

            self._head = QWidget(self)
            l_head = QHBoxLayout(self._head)
            l_head.setContentsMargins (*self._head_margins)

            if self._switchable:
                CheckBox (l_head, fontSize=size.HEADER, text=self.title_text(),
                        get=lambda: self.switched_on, set=self.set_switched_on)
            else: 
                Label (l_head, fontSize=size.HEADER, get=self.title_text)
            l_head.setStretch (0,1)

            self._add_to_header_layout (l_head)     # optional individual widgets
 
        # inital content panel content - layout in >init  

        self._panel  = QWidget() 
        self._initial_layout = layout 

        if not lazy_layout: 
            self._set_panel_layout () 

        # main layout with title and panel 

        l_main   = QVBoxLayout()
        if self._head: 
            l_main.addWidget (self._head)
        l_main.addWidget (self._panel, stretch=2)
        l_main.setContentsMargins (QMargins(*self._main_margins))  
        l_main.setSpacing(2)
        self.setLayout (l_main)

        # initial switch state 
        self.set_switched_on (self._switched_on, silent=True)

        # initial enabled/disabled state
        if self._isDisabled: 
            self.refresh_widgets (self._isDisabled) 

        # initial visibility 
        if not self.shouldBe_visible:         
            self.setVisible (False)             # setVisible(True) results in a dummy window on startup 


    def title_text (self) -> str: 
        """ returns text of title - default self.name"""
        # can be overwritten 
        return self.name 


    @property 
    def switched_on (self) -> bool:
        """ True if self is switched on"""
        return self._switched_on
    
    def set_switched_on (self, aBool : bool, silent=False):
        """ switch on/off 
            - optional hide main panel 
            - option callback on_switched
        """
        self._switched_on = aBool is True 
        
        if self._hide_switched:
            self._panel.setVisible (self.switched_on)

            if self.switched_on:
                Widget._set_height (self, self._height)
            else: 
                Widget._set_height (self, 40)

        # refresh also header checkbox
        for w in self.header_widgets:
            w.refresh ()

        if not silent and callable (self._on_switched):                                          # set by checkbox - callback to Diagram_Item 
            self._on_switched (self._switched_on)


    @property
    def widgets (self) -> list[Widget]:
        """ list of Widgets defined in self panel area"""

        return self.widgets_of_layout (self._panel.layout())


    @property
    def header_widgets (self) -> list[Widget]:
        """ list of widgets defined in self header area"""

        if self._head: 
            return self.widgets_of_layout (self._head.layout())
        else: 
            return []
 

    def refresh(self, reinit_layout=False):
        """ refreshes all Widgets on self """

        hide = not self.shouldBe_visible and     self.isVisible()
        show =     self.shouldBe_visible and not self.isVisible()

        # hide / show self 

        if hide or show: 
            self.setVisible (self.shouldBe_visible)
            logger.debug (f"{self} - setVisible ({self.shouldBe_visible})")

        if self.isVisible():

            # reinit layout 
            if show or reinit_layout:
                self._set_panel_layout  () 
                logger.debug (f"{self} - refresh - reinit_layout: {reinit_layout} ")

            # refresh widgets of self only if visible 
            self.refresh_widgets (self._isDisabled)
            logger.debug (f"{self} - refresh widgets   disable: {self._isDisabled}")


    def refresh_widgets (self, disable : bool):
        """ enable / disable all widgets of self - except Labels (color!) """

        w : Widget
        for w in self.widgets:
            w.refresh (disable=disable)

        # refresh also header 
        for w in self.header_widgets:
            w.refresh (disable=disable)


    def _set_panel_layout (self):
        """ 
        Set layout of self._panel   
            - typically defined in subclass in _init_layout
            - or as init argument 'layout'"""

        # remove and rebuild only in case of _init_layout done in a subclass

        if self._initial_layout:

            layout = self._initial_layout

        else:
            if self._panel.layout() is not None:
                self._clear_existing_panel_layout ()

            if self.shouldBe_visible:
                layout = self._init_layout()        # subclass will create layout 
            else: 
                # if the panel shouldn't be visible repalce the normal panel layout
                # with a dummy, so get/set of the normal panel widgets won't get trouble 
                layout = QVBoxLayout()               
                wdt = QLabel ("This shouldn't be visible")
                layout.addWidget (wdt)

        layout.setContentsMargins (QMargins(*self._panel_margins))   # inset left 
        layout.setSpacing(2)
        self._panel.setLayout (layout)


    def _clear_existing_panel_layout(self):
        """ removes all items from the existing panel layout"""
        layout = self._panel.layout()
        if layout is not None: 
            while layout.count():
                child = layout.takeAt(0)
                childWidget = child.widget()
                if childWidget:                
                    childWidget.deleteLater() # sip.delete (childWidget)
                sip.delete (child)
                # if childWidget:
                #     childWidget.disconnect()
                #     childWidget.setParent(None)
                #     sip.delete (childWidget)
                #     # childWidget.deleteLater()                             # will create ghost widget due to async event loop
            logger.debug (f"{self} - clear layout ")

            # finally remove self 
            sip.delete (self._panel.layout())


    def _init_layout(self) -> QLayout:
        """ init and return main layout"""

        # to be implemented by sub class
        pass


    def _add_to_header_layout(self, l_head : QHBoxLayout):
        """ add Widgets to header layout"""

        # to be implemented by sub class
        pass



# ------------ MessageBox  -----------------------------------


class MessageBox (QMessageBox):
    """ 
    Subclass of QMessagebox 
        - new default icons 
        - more width and height 
    """

    _min_width  = 250
    _min_height = 80 

    def __init__(self, parent: object, 
                 title : str, 
                 text : str, 
                 icon: Icon, 
                 min_width=None,                            # width of text widget 
                 min_height=None):                          # height of text widget 
        super().__init__(parent)

        # set properties 

        self.setWindowTitle (title)
        self.setText (text)

        # set icon 

        if isinstance (icon, Icon):
            pixmap = icon.pixmap((QSize(32, 32)))
            self.setIconPixmap (pixmap)

        # set width and height 
        #   size of QMessageBox must be set via layout - which is a bit hacky 

        layout : QGridLayout = self.layout()
        if isinstance (layout, QGridLayout):

            cols = layout.columnCount()
            if cols > 1:
                min_width = min_width if min_width is not None else self._min_width
                # set minimum width of last column (which should be text) 
                layout.setColumnMinimumWidth (cols-1,min_width)
                layout.setColumnStretch (cols-1,3)
                # set minimum width of first column (which should be icon) 
                layout.setColumnMinimumWidth (0,60)
                item = layout.itemAtPosition (0,1)
                item.setAlignment (Qt.AlignmentFlag.AlignVCenter )

                # increase right margin 
                left, top, right, bottom = layout.getContentsMargins()
                layout.setContentsMargins (left, top, right+20, bottom)

            rows = layout.columnCount()
            if rows > 1:
                min_height = min_height if min_height is not None else self._min_height
                # set minimum widthof last column (which should be text) 
                layout.setRowMinimumHeight (0,min_height)
                # set center alignment of icon 
                item = layout.itemAtPosition (0,0)
                item.setAlignment (Qt.AlignmentFlag.AlignCenter )

        if os.name == 'posix':
            # strange posix - wayland bug 
            #   With first call of MessageBox the box is positioned at 0,0 -> manually move to the center of parent 
            point = parent.geometry().center()
            x = point.x() - 200
            y = point.y() - 100
            self.move(x,y)
 

    @staticmethod
    def success (parent: object, title : str, text : str, min_width=None, min_height=None):
        """ success message with Ok button"""

        msg = MessageBox (parent, title, text, Icon (Icon.SUCCESS), min_width=min_width, min_height= min_height)
        msg.exec()


    @staticmethod
    def info (parent: object, title : str, text : str, min_width=None, min_height=None):
        """ info message with Ok button"""

        msg = MessageBox (parent, title, text, Icon (Icon.INFO), min_width=min_width, min_height= min_height)
        msg.exec()


    @staticmethod
    def error (parent: object, title : str, text : str, min_width=None, min_height=None):
        """ critical message with Ok button"""

        msg = MessageBox (parent, title, text, Icon (Icon.ERROR), min_width=min_width, min_height= min_height)
        msg.exec()


    @staticmethod
    def confirm (parent: object, title : str, text : str, min_width=None):
        """ confirmation with Ok and Cancel"""

        msg = MessageBox (parent, title, text, Icon (Icon.INFO), min_width=min_width)

        msg.setStandardButtons(QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)
        msg.setDefaultButton  (QMessageBox.StandardButton.Ok)

        return msg.exec()


    @staticmethod
    def warning (parent: object, title : str, text : str, min_width=None):
        """ warning with Ok and Cancel"""

        msg = MessageBox (parent, title, text, Icon (Icon.WARNING), min_width=min_width)

        msg.setStandardButtons(QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)
        msg.setDefaultButton  (QMessageBox.StandardButton.Ok)

        return msg.exec()


    @staticmethod
    def save (parent: object, title : str, text : str, min_width=None, buttons=None):
        """ ask to save or discard - returns QMessageBox.StandardButton"""

        msg = MessageBox (parent, title, text, Icon (Icon.WARNING), min_width=min_width)

        if buttons is None: 
            buttons = QMessageBox.StandardButton.Save | QMessageBox.StandardButton.Discard | QMessageBox.StandardButton.Cancel 
        msg.setStandardButtons(buttons)
        msg.setDefaultButton  (QMessageBox.StandardButton.Save)

        return msg.exec()



# ------------ Toaster  -----------------------------------


class Toaster (QFrame):

    """ 
    Show a little notfication toaster 

    based on https://stackoverflow.com/questions/59251823/is-there-an-equivalent-of-toastr-for-pyqt
    """

    def __init__(self, *args, **kwargs):

        super (Toaster, self).__init__(*args, **kwargs)
        QHBoxLayout(self)

        self.label = None

        self.setSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)

        #                 border: 1px solid blue;
        # background: palette(window);

        # self.setStyleSheet('''
        #     QToaster {
        #         border-radius: 2px; 
        #     }
        # ''')
        # alternatively:

        self.setAutoFillBackground(True)
        # set_background (self, color="red")
        # self.setFrameShape(self.Box)

        self.timer = QTimer(singleShot=True, timeout=self.hide)

        if self.parent():
            self.opacityEffect = QGraphicsOpacityEffect(opacity=0)
            self.setGraphicsEffect (self.opacityEffect)
            self.opacityAni = QPropertyAnimation (self.opacityEffect, b'opacity')
            # we have a parent, install an eventFilter so that when it's resized
            # the notification will be correctly moved to the right corner
            self.parent().installEventFilter(self)
        else:
            # there's no parent, use the window opacity property, assuming that
            # the window manager supports it; if it doesn't, this won'd do
            # anything (besides making the hiding a bit longer by half a second)
            self.opacityAni = QPropertyAnimation(self, b'windowOpacity')
        self.opacityAni.setStartValue(0.)
        self.opacityAni.setEndValue(1.)
        self.opacityAni.setDuration(100)
        self.opacityAni.finished.connect(self.checkClosed)

        self.corner = Qt.Corner.TopLeftCorner
        self.margin = QMargins(10, 10, 10, 10)


    def checkClosed(self):
        # if we have been fading out, we're closing the notification
        if self.opacityAni.direction() == QAbstractAnimation.Direction.Backward:
            self.close()

    def restore(self):
        # this is a "helper function", that can be called from mouseEnterEvent
        # and when the parent widget is resized. We will not close the
        # notification if the mouse is in or the parent is resized
        self.timer.stop()
        # also, stop the animation if it's fading out...
        self.opacityAni.stop()
        # ...and restore the opacity
        if self.parent():
            self.opacityEffect.setOpacity(1)
        else:
            self.setWindowOpacity(1)

    def hide(self):
        # start hiding
        self.opacityAni.setDirection(QAbstractAnimation.Direction.Backward)
        self.opacityAni.setDuration(300)
        self.opacityAni.start()

    def eventFilter(self, source, event : QEvent):
        if source == self.parent() and event.type() == QEvent.Type.Resize:

            self.opacityAni.stop()

            parentRect : QRectF = self.parent().rect()
            geo = self.geometry()

            corner = self.corner
            margin = self.margin

            if corner == Qt.Corner.TopLeftCorner:
                geo.moveTopLeft(parentRect.topLeft() + QPoint(margin.left(), margin.top()))
            elif corner == Qt.Corner.TopRightCorner:
                geo.moveTopRight(parentRect.topRight() + QPoint(-margin.right(), margin.top()))
            elif corner == Qt.Corner.BottomRightCorner:
                geo.moveBottomRight(parentRect.bottomRight() + QPoint(-margin.right(), -margin.bottom()))
            else:
                geo.moveBottomLeft(parentRect.bottomLeft() + QPoint(margin.left(), -margin.bottom()))

            self.setGeometry(geo)
            self.restore()
            self.timer.start()
        return super(Toaster, self).eventFilter(source, event)

    def enterEvent(self, event):
        self.restore()

    def leaveEvent(self, event):
        self.timer.start()

    def closeEvent(self, event):
        # we don't need the notification anymore, delete it!
        self.deleteLater()

    def resizeEvent(self, event):
        super(Toaster, self).resizeEvent(event)
        # if you don't set a stylesheet, you don't need any of the following!
        if not self.parent():
            # there's no parent, so we need to update the mask
            path = QPainterPath()
            path.addRoundedRect(QRectF (self.rect()).translated(-.5, -.5), 4, 4)
            self.setMask(QRegion(path.toFillPolygon(QTransform()).toPolygon()))
        else:
            self.clearMask()

    @staticmethod
    def showMessage(parent, message, 
                    corner=Qt.Corner.BottomLeftCorner, 
                    margin=QMargins(10, 10, 10, 10), 
                    toast_style = style.HINT,
                    duration=2000, parentWindow=False):
        """
        show a toaster for a while

        Args:
            parent : parent Toasters position is relative 
            message: message to show 
            corner : position in parents window 
            margin : offset of position
            style : color style 
            duration: duration of message 
            parentWindow : take parent window of parent
        """

        if parent and parentWindow:
            parent = parent.window()

        if isinstance (margin, int):
            margin = QMargins(margin, margin, margin, margin)

        if toast_style is None: 
            toast_style = style.HINT

        self = Toaster(parent)
        parentRect : QRectF = parent.rect()

        self.timer.setInterval(duration)

        self.label = QLabel(message)
        self.layout().addWidget(self.label)
        self.layout().setContentsMargins (QMargins(30, 7, 30, 7))

        # set background color style 

        if toast_style in [style.WARNING, style.ERROR, style.COMMENT, style.GOOD, style.HINT]:

            palette : QPalette = self.palette()
            index = Widget.LIGHT_INDEX if Widget.light_mode else Widget.DARK_INDEX
            color = QColor (toast_style.value[index])

            # adapt color 

            h,s,l,a = color.getHsl()
            if Widget.light_mode:
                s = int(s*0.8)                                  # less saturation (more gray) 
                l = int(l*1.4)                                  # lighter 
            else:
                s = int(s*0.8)                                  # less saturation (more gray) 
                l = int(l*0.5)                                  # darker 
            color.setHsl (h, s, l, a)

            palette.setColor(QPalette.ColorRole.Window, color)
            self.setPalette(palette)

        # start hide timer 

        self.timer.start()

        # raise the widget and adjust its size to the minimum

        self.raise_()
        self.adjustSize()

        self.corner = corner
        self.margin = margin

        geo = self.geometry()

        # now the widget should have the correct size hints, let's move it to the right place
        
        if corner == Qt.Corner.TopLeftCorner:
            geo.moveTopLeft(parentRect.topLeft() + QPoint(margin.left(), margin.top()))
        elif corner == Qt.Corner.TopRightCorner:
            geo.moveTopRight(parentRect.topRight() + QPoint(-margin.right(), margin.top()))
        elif corner == Qt.Corner.BottomRightCorner:
            geo.moveBottomRight(parentRect.bottomRight() + QPoint(-margin.right(), -margin.bottom()))
        else:
            geo.moveBottomLeft(parentRect.bottomLeft() + QPoint(margin.left(), -margin.bottom()))

        self.setGeometry(geo)
        self.show()
        self.opacityAni.start()



# ------------ Dialog  -----------------------------------

class Dialog (QDialog):
    """
    Abstract super class for modal top windows with action buttons at bottom.
    Extends QDialog with a dataObject (via 'getter') and common background

    Args:
        parent: parent widget self (position) should be related
        getter: bound method for data object 
        width : overwrite default width of self 
        height: overwrite default height of self 
        dx    : move top left corner dx pixel to the right of parent bottom left
        dy    : move top left corner dy pixel down of parent bottom left
    Returns:
        _type_: _description_
    """

    name = "Dialog"             # will be title 

    _width  = None
    _height = None 


    def __init__(self,  
                 parent : QWidget = None,
                #  flags : Qt.WindowType = None, 
                 getter = None, 
                 width  : int | None =None, 
                 height : int | None =None, 
                 dialogPos : tuple = (0.4,0.5),          # reference point for positioning - (0,0) is upper, left corner
                 parentPos : tuple = (0.3,0.6),          # position anchor in parent - (0,0) is upper, left corner
                 dx : int | None = None, 
                 dy : int | None = None, 
                 title=None, 
                 flags=Qt.WindowType.Dialog, **kwargs):
        super().__init__(parent, flags=flags, **kwargs)

        self._parent = parent
        self._getter = getter

        self._dataObject_copy = copy (self.dataObject)

        if width is not None: 
            self._width = width
        if height is not None: 
            self._height = height

        # set width and height 

        Widget._set_width  (self, self._width)
        Widget._set_height (self, self._height)

        # move self relative to parent bottom left
        #       calc from the relative parentPos and dialogPos arguments the absolute positioning

        parent_topLeft     = self._parent.mapToGlobal (self._parent.rect().topLeft())
        parent_bottomRight = self._parent.mapToGlobal (self._parent.rect().bottomRight())
        parent_width  = parent_bottomRight.x() - parent_topLeft.x()
        parent_height = parent_bottomRight.y() - parent_topLeft.y()

        parentPos_rel = QPoint (int(parent_width * parentPos[0]), int(parent_height * parentPos[1]))
        parentPos_abs = parent_topLeft + parentPos_rel

        self_width  = self.rect().width()
        self_height = self.rect().height()
        selfPos_rel = QPoint (int(self_width * dialogPos[0]), int(self_height * dialogPos[1]))

        if dx is not None and dy is not None:
            dxy_rel = QPoint (int(dx), int(dy))
        else:
            dxy_rel = QPoint (0,0)

        self.move (parentPos_abs - selfPos_rel + dxy_rel)

        # title of dialog 

        if title is not None: 
            self.name = title 
        self.setWindowTitle (self.name)

        # enable custom window hint, disable (but not hide) close button
        # self.setWindowFlag(Qt.WindowType.WindowCloseButtonHint, False)

        # inital content panel content - layout in _init_layout()  

        self._panel = QWidget () 
        self.set_background_color (darker_factor=105)

        l_panel = self._init_layout()                               # subclass will create layout 
        l_panel.setContentsMargins (QMargins(15, 10, 15, 10))       # inset left 
        self._panel.setLayout (l_panel)

        # Qt buttonBox at footer

        button_box = self._button_box()
        if button_box:
            l_button = QHBoxLayout()
            l_button.addWidget(self._button_box())
            l_button.setContentsMargins (QMargins(5, 0, 25, 0))
 
        # main layout with title and panel 

        l_main   = QVBoxLayout()
        l_main.addWidget (self._panel, stretch=1)
        if button_box:
            l_main.addLayout (l_button)
        l_main.setContentsMargins (QMargins(5, 5, 5, 15))
        l_main.setSpacing(15)
        self.setLayout (l_main)

        # connect to change signal of widget 
        
        for w in self.widgets:
            w.sig_changed.connect (self._on_widget_changed)


    def set_background_color (self, darker_factor : int | None = None,
                                    color : QColor | int | None  = None,
                                    alpha : float | None = None):
        """ 
        Set background color of a QWidget either by
            - darker_factor > 100  
            - color: QColor or string for new color
            - alpha: transparency 0..1 
        """
        set_background (self._panel, darker_factor=darker_factor, color=color, alpha=alpha)


    def _init_layout(self) -> QLayout:
        """ init and return main layout"""

        # to be implemented by sub class
        return QVBoxLayout ()

    def _button_box (self) -> QDialogButtonBox:
        """ returns the QButtonBox with the buttons of self"""
        buttons = QDialogButtonBox.StandardButton.Ok | \
                  QDialogButtonBox.StandardButton.Cancel
        buttonBox = QDialogButtonBox(buttons)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)
        return buttonBox 

    def _QButtons (self):
        """return QButtons enum for button box at footer """
        # to overload 
        return QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel


    @property
    def dataObject (self): 
        """ dataObject the dialog got from parent """
        # to be overloaded - or implemented with semantic name 
        if callable(self._getter):
            return self._getter()
        else: 
            return self._getter

    @property
    def dataObject_copy (self): 
        """ shallow copy of dataObject"""
        # to be overloaded - or implemented with semantic name 
        return self._dataObject_copy

    @property
    def widgets (self) -> list[Widget]:
        """ list of widgets defined in self """
        return Panel_Abstract.widgets_of_layout (self._panel.layout())


    def _on_widget_changed (self,*_):
        """ slot for change of widgets"""
        # to be overloaded 
        pass


    def button_clicked (self, aButton): 
        """ slot for button of buttonbox clicked. Can be overriden"""
        pass


    def refresh(self, disable=None):
        """ refreshes all Widgets on self """

        for w in self.widgets:
            w.refresh(disable=disable)
        logger.debug (f"{self} - refresh")


    @override
    def reject (self): 
        """ handle reject (close) actions"""
        # to override 
        super().reject()


# ------------------------------------------------------------------------------
# ------------ test functions - to activate  -----------------------------------
# ------------------------------------------------------------------------------

# class Test_Panel (Edit_Panel):

#     name = "Airfoil Data"
#     _width  = 200
#     _height = (100, None)

#     def _init_layout (self)-> QLayout: 
#         l = QGridLayout()
#         Label  (l,0,0,get="Ein Label")
#         Button (l,2,0, text= "Open dialog", set=self._open_dialog)
#         l.setRowStretch (0,1)
#         return l 

#     def _open_dialog (self):

#         dlg = Test_Dialog(parent=self, title="mein Titel", height=400, width=600)
#         dlg.exec ()


# class Test_Dialog (Dialog):

#     def _init_layout(self) -> QLayout:
#         l = QGridLayout()
#         Label  (l,0,0,get="Hallo Dialog")
#         SpaceR (l,1)
#         Label  (l,2,0,get="ganz unten")
#         l.setRowStretch (1,1)
#         return l 


# class Test_Panel2 (Edit_Panel):

#     name   = "Curvature"
#     _width  = 200
#     _height = (200, None)

#     def _init_layout (self) -> QLayout: 
#         l = QGridLayout()
#         r = 0 
#         Label  (l,0,0, colSpan=2, get="Ein Läbele extra lang")
#         r += 1
#         Label  (l,r,0,get="Weight", width=70)
#         FieldI (l,r,1,get=15, lim=(0,100), unit="kg", step=1, width=(70,90)) # (70, 90)
#         r += 1
#         Label  (l,r,0,get="Span")
#         FieldI (l,r,1,get="2980", lim=(0,9999), unit="mm", step=10, width=(100, None))
#         r+= 1
#         SpaceR (l,r)
#         l.setColumnStretch (1,3)
#         l.setColumnStretch (2,1)
#         l.setRowStretch (r, 1)
#         return l 


# class Test_Panels (QMainWindow):

#     def __init__(self):
#         super().__init__()

#         self.setWindowTitle('Test Edit_Panel')
#         self.setMinimumSize(QSize(400, 200))

#         air_panel  = Test_Panel  (self) 
#         curv_panel = Test_Panel2 (self, switchable=True,hide_switched=True) 

#         container = QWidget()
#         l = QHBoxLayout ()
#         l.addWidget (air_panel) 
#         l.addWidget (curv_panel)
#         container.setLayout (l)  

#         self.setCentralWidget(container)


# if __name__ == "__main__":

#     from PyQt6.QtWidgets import QApplication, QMainWindow, QWidget, QGridLayout
#     # logging.basicConfig(level=logging.DEBUG)
 
#     app = QApplication([])
#     app.setStyle('fusion')
#     app.setStyleSheet ("QWidget { font-family: 'Segoe UI' }")

#     w = Test_Panels()
#     w.show()
#     app.exec() 



#!/usr/bin/env pythonbutton_color
# -*- coding: utf-8 -*-

"""  

All abstract diagram items to build a complete diagram view 

"""

import logging

from typing             import override, Type

from PyQt6.QtCore       import QSize, QMargins, pyqtSignal, QTimer
from PyQt6.QtWidgets    import QLayout, QGridLayout, QVBoxLayout, QHBoxLayout, QGraphicsGridLayout
from PyQt6.QtWidgets    import QWidget
from PyQt6.QtGui        import QPalette, QIcon, QColor

import pyqtgraph as pg                          # import PyQtGraph after PyQt6
from pyqtgraph          import icons

from base.common_utils  import *
from base.panels        import Edit_Panel, Container_Panel
from base.widgets       import ToolButton, Icon, Widget
from base.artist        import Artist

logger = logging.getLogger(__name__)
# logger.setLevel(logging.WARNING)



class Diagram (QWidget):
    """ 
    Abstract container panel/widget which may hold
     - serveral PlotItems 
     - each having optional view panel sections (to the left) 

    see: https://pyqtgraph.readthedocs.io/en/latest/getting_started/plotting.html

    """

    _width  = (200, None)                # (min,max) 
    _height = (100, None)                # (min,max)

    name   = "My Diagram"               # will be shown in Tabs 


    def __init__(self, parent, 
                 getter = None, 
                 width=None, 
                 height=None, 
                 **kwargs):
        super().__init__(parent, **kwargs)

        self._getter = getter
        self._myApp  = parent
        self._section_panel = None 

        # set width and height 

        if width is not None: 
            self._width = width
        if height is not None: 
            self._height = height

        Widget._set_width  (self, self._width)
        Widget._set_height (self, self._height)

        # create graphics widget 

        self._graph_widget = pg.GraphicsLayoutWidget (parent=self, show=True, size=None, title=None)
        self.graph_layout.setContentsMargins (20,20,20,10)  # default margins

        # remove default context menu of scene

        self._graph_widget.ci.scene().contextMenu = []


        # create all plot items and setup layout with them  

        self.create_diagram_items () 

        #  add a message view box at bottom   

        self._message_vb = pg.ViewBox()
        nrows = self.graph_layout.rowCount()
        self.graph_layout.addItem (self._message_vb, nrows, 0)
        self._message_vb.hide()
    

        # create optional view panel add the left side 

        self._viewPanel   = None
        self.create_view_panel ()

        # build layout with view panel and graphics

        l_main = QHBoxLayout()
        l_main.setContentsMargins (QMargins(0, 0, 0, 0))
        # l_main.setSpacing (0)

        if self._viewPanel is not None: 
            l_main.addWidget (self._viewPanel)
        l_main.addWidget (self._graph_widget, stretch=10)

        self.setLayout (l_main)


    def __repr__(self) -> str:
        # overwritten to get a nice print string 
        text = '' 
        return f"<{type(self).__name__}{text}>"


    def data_list (self): 
        # to be overloaded - or implemented with semantic name   

        if callable(self._getter):
            obj = self._getter()
        else: 
            obj = self._getter     
        return obj if isinstance (obj, list) else [obj]


    def _get_items (self, item_classes : Type['Diagram_Item'] | list[type['Diagram_Item']]) -> list['Diagram_Item']:
        """get Diagram Items of self having class name(s)

        Args:
            items: class or list of class of Diagram Items to retrieve
        Returns:
            List of Item with this classes
        """
        look_for = [item_classes] if not isinstance (item_classes,list) else item_classes
        result = []
        for item in self.diagram_items:
            if item.__class__ in look_for:
                result.append(item)
        return result 


    def _get_first_item (self, item_class : Type['Diagram_Item']) -> list['Diagram_Item']:
        """get Diagram Items of self having class name(s)"""
        items = self._get_items (item_class)
        if items:
            return items [0]
        else: 
            return None
        

    def _show_section_and_item (self, item_class : Type['Diagram_Item'], show = True) -> list['Diagram_Item']:
        """show view_panel section and the diagram item """
        for item in self._get_items (item_class):
            item.section_panel.set_switched_on (show)


    def _get_artist (self, artists : Type[Artist] | list[type[Artist]]) -> list[Artist]:
        """get artists of all my items having class name(s)

        Args:
            artists: class or list of class of Artists to retrieve
        Returns:
            List of artists with this classes
        """
        result = []
        for item in self.diagram_items: 
            result.extend (item._get_artist (artists))
        return result 


    def _show_artist (self, artist_class : Type[Artist], show = True, refresh = True):
        """show on/off of artist having artist_class name """

        for item in self.diagram_items:
            item._show_artist (artist_class, show, refresh=refresh)


    @property
    def myApp (self): 
        return self._myApp
    
    @property
    def diagram_items (self) -> list['Diagram_Item']:
        """ list of my diagram items """
        items = []
        for i in range (self.graph_layout.count()):
            item = self.graph_layout.itemAt(i)
            if isinstance (item, Diagram_Item):
                items.append (item)
        return items

    @property
    def graph_layout (self) -> QGraphicsGridLayout:
        """ the graphics grid layout of self"""
        return self._graph_widget.ci.layout

    @property
    def section_panel (self) -> Edit_Panel | None:
        """ small section panel representing self in view panel"""
        # overload for constructor 
        return self._section_panel
    

    def refresh(self, also_viewRange=True): 
        """ 
        refresh all childs (Diagram_Items) of self
        Args:
           also_viewRange: also re-init viewRange     
        """

        if self.isVisible():

            logger.debug (f"{str(self)} refresh")

            if self._viewPanel:
                self._viewPanel.refresh()                                       # first view panel as diagram settings could change

            item : Diagram_Item
            for item in self.diagram_items:
                if item.isVisible(): 
                    item.refresh()
                if also_viewRange:                                              # also setup view range if not visible
                    item.setup_viewRange()  


    @override
    def showEvent (self, ev):
        """ QShowEvent - self becomes visible  - Tab switch or first time"""

        # refresh plotItems - if it is first time they are intially plotted

        item : Diagram_Item
        for item in self.diagram_items:
            if item.isVisible(): 
                item.refresh() 

        # refresh all panels on viewPanel 

        if self._viewPanel:
            self._viewPanel.refresh()

        super().showEvent (ev) 



    def create_diagram_items ():
        """ create all plot Items and add them to the layout """
        # to be overlaoded 
        # like ... 
        #   item = Airfoil_Diagram_Item (self, getter=self.airfoil)
        #   self._add_item (item, 0, 0)
        pass


    def _add_item (self, anItem: 'Diagram_Item', row, col, rowspan=1, colspan=1):
        """ adds a diagram item to self graphic layout """

        self._graph_widget.addItem (anItem, row, col, rowspan=rowspan, colspan=colspan)

        anItem.sig_visible.connect (self._on_item_visible)


    def create_view_panel (self):
        """ 
        creates a view panel to the left of at least one diagram item 
        has a section_panel
        """

        # build side view panel with the section panels 

        layout = QVBoxLayout()
        layout.setContentsMargins (QMargins(0, 0, 0, 0)) 
        for item in self.diagram_items:
            if item.section_panel is not None: 
                layout.addWidget (item.section_panel,stretch=1)

        # add section panel of self (master) 

        if self.section_panel is not None: layout.addWidget (self.section_panel,stretch=0)
        
        layout.addStretch (1)

        self._viewPanel = Container_Panel()
        self._viewPanel.setMinimumWidth(180)
        self._viewPanel.setMaximumWidth(250)
        self._viewPanel.setLayout (layout)


    def _on_item_visible (self, aBool):
        """ slot to handle actions when switch on/off of diagram items"""

        nItems = len([item for item in self.diagram_items if item.isVisible()])

        if nItems == 0: 
            self._message_show ("No diagram items selected ")
        # the GraphicsLayout gets confused, if all items were switched off
        #     and then switched on again - recalc layout
        if aBool: 
            if nItems > 0:
                self._message_clear()


    def _message_show (self, aMessage):
        """ shows aMessage in the middle of the diagram """
        text = pg.TextItem(aMessage, anchor=(0.5,0))
        text.setPos(0, 0)
        self._message_vb.addItem (text)
        self._message_vb.show()


    def _message_clear (self):
        """ shows aMessage in the middle of the diagram """
        self._message_vb.clear()
        self._message_vb.hide()




class Diagram_Item (pg.PlotItem):
    """ 
    Abstract PlotItem  

    see: https://pyqtgraph.readthedocs.io/en/latest/getting_started/plotting.html

    """

    name        = "Abstract Diagram_Item"               # used for link and section header
    title       = "The Title"                           # title of diagram item
    subtitle    = "my subtitle"                         # optional subtitle 

    min_width   = 500                                   # min size needed - see below 
    min_height  = 150 

    # Signals 

    sig_visible = pyqtSignal(bool)                      # when self is set to show/hide 


    def __init__(self, parent, 
                 getter = None, 
                 show = True,                           # show initially 
                 **kwargs):

        super().__init__(name=self.name,                # to link view boxes 
                         **kwargs)

        self._parent : Diagram = parent 
        self._getter = getter
        self._show   = show 

        self._section_panel = None                      # view section to the left 
        self._artists : list [Artist] = []              # list of my artists

        self._title_item = None                         # LabelItem of the diagram title
        self._subtitle_item = None  

        self._help_messages         = {}                # current help messages which are shown 
        self._help_messages_shown   = {}                # all help messages shown up to now 
        self._help_message_items    = {}

        # setup additional, optional button to reset view range 

        self._vb_state_changed = False 
        self._resetBtn = None
        if self.has_reset_button:
            ico = Icon (Icon.RESETVIEW,light_mode = False)
            self._resetBtn = pg.ButtonItem(pixmap=ico.pixmap(QSize(52,52)), width=16, parentItem=self)
            self._resetBtn.mode = 'auto'
            self._resetBtn.clicked.connect(self._resetBtn_clicked)
            timer = QTimer()                                
            timer.singleShot(50, self._reset_prepare)           # delayed when all initial drawing done 

        # setup item to print coordinates 

        self._coordItem = pg.LabelItem("", color=QColor(Artist.COLOR_LEGEND), size=f"{Artist.SIZE_NORMAL}pt", justify="left")  
        self._coordItem.setParentItem(self)  
        self._coordItem.anchor(parentPos=(0,1), itemPos=(0.0,0.0), offset=(45, -20))                       

        # set margins (inset) of self - ensure some space for coordinates

        self.setContentsMargins ( 10,20,10,20)

        # PlotItem needs a some size so that inital boundingBox calculation having pixel calculation work properly
        #   - see pixelVectors  (important for clickable and movable (mousesize))
        
        self.setMinimumSize (self.min_width, self.min_height)

        # setup artists - must be override

        self.setup_artists ()

        for artist in self._artists:
            artist.sig_help_message.connect (self._on_help_message)

        # setup view range (must be override) is done in refresh() - 

        self._viewRange_set = False

        # setup axis - can be override
        
        self.setup_axis()

        # no pyqtgraph context menu  - no view box context menu , no plot item 

        self.setMenuEnabled(enableMenu=False, enableViewBoxMenu=False)

        # initial show or hide - use super() - avoid refresh
 
        super().setVisible (show) 


    @override
    def __repr__(self) -> str:
        text = '' 
        return f"<{type(self).__name__}{text}>"


    def _add_artist (self, artist : Artist):
        """ add a new artist to my list"""  
    
        self._artists.append (artist)
     
    def _get_artist (self, artists : Type[Artist] | list[type[Artist]]) -> list[Artist]:
        """get my artists having class name(s)

        Args:
            artists: class or list of class of Artists to retrieve
        Returns:
            List of artists with this classes
        """
        look_for = [artists] if not isinstance (artists,list) else artists
        result = []
        for artist in self._artists:
            if artist.__class__ in look_for:
                result.append(artist)
        return result 


    def _show_artist (self, artist_class : Type[Artist], show : bool = True, refresh=True):
        """show on/off of artist having artist_class name """

        # logger.debug (f"{self} show artists: {show} - is visible: {self.isVisible()}")
        for artist in self._get_artist (artist_class):
            artist.set_show (show, refresh=refresh)            


    def _on_help_message (self, aArtist :Artist | None, aMessage: str | None):
        """ slot for help message signal of an artist. show it"""

        # if aArtist: 
        #     logger.warning (f"{self} on_help of {aArtist} of item: {aArtist._pi} with message: {aMessage}")
        # else: 
        #     logger.warning (f"{self} on_help  -> refresh ")

        # remove all existing message item 
        for item in self._help_message_items.values():
            self.scene().removeItem (item)                             # was added directly to the scene via setParentItem
        self._help_message_items = {}

        # add / remove message of artist to my list 
        if aArtist:                                                    # could be called without artist -> refresh only 
            if not aMessage:
                self._help_messages.pop (aArtist, None)                # remove existing message of this artist
            else:
                if aArtist in self._help_messages_shown:
                    self._help_messages.pop (aArtist, None) 
                else:   
                    self._help_messages[aArtist] = aMessage            # add this message 
                    self._help_messages_shown[aArtist] = aMessage      # add to the list already shown           


        # plot all current messages 
        i = 0 
        for artist, message in self._help_messages.items():
            parentPos = (0.3,0.03)                                      # parent x starts at PlotItem (including axis)       
            itemPos   = (0.0,0.0)
            y_offset  = i * 20

            p1 = pg.LabelItem(f"- {message}", color=Artist.COLOR_HELP, size=f"{Artist.SIZE_NORMAL}pt")    

            p1.setParentItem(self)                                      # add to self (Diagram Item) for absolute position 
            p1.anchor(itemPos=itemPos, parentPos=parentPos, offset=(0,y_offset))
            p1.setZValue(5)

            self._help_message_items[artist] = p1
            i +=1


    @override
    def setVisible (self, aBool):
        """ Qt overloaded to signal parent """
        super().setVisible (aBool)

        logger.debug  (f"{self} - setVisible {aBool}")

        self.set_show (aBool)

        self.sig_visible.emit (aBool)


    def set_show (self, aBool):
        """ switch on/off artists of self"""
        # to override
        self._show = aBool

        if aBool: 
            self.refresh_artists ()
            self.setup_viewRange()   
            self._viewRange_set = True

            self.plot_title ()


    @override
    def close (self):
        # PlotItem override to remove button
        if self._resetBtn is not None:
            self._resetBtn.setParent (None) 
            self._resetBtn = None
        super().close()


    @override
    def hoverEvent(self, ev):
        """ overridden to show coordinates of mouse cursor"""
        super().hoverEvent (ev)

        if self.mouseHovering and not self.buttonsHidden:
            pos : pg.Point = self.viewBox.mapSceneToView(ev.scenePos())

            if self.viewBox.viewRect().contains(pos):
                self._coordItem.show()
                self._coordItem.setText (self._coord_as_text (pos))
            else:
                self._coordItem.hide()


    def _coord_as_text (self, pos : pg.Point) -> str:
        """ current mouse coordinates as formatted text string """

        viewRange = self.viewBox.viewRange()
        x_range = viewRange [0]
        y_range = viewRange [1]
        x_size  = abs (x_range[1] - x_range[0])
        y_size  = abs (y_range[1] - y_range[0])

        return f"x {self._format_coord(pos.x(),x_size)}&nbsp;&nbsp;y {self._format_coord(pos.y(),y_size)}"

    def _format_coord (self, coord: float, range : float) -> str:
        """ format a coordinate number depending on view range size"""
        if range < 0.01:
            dec = 5
        elif range < 0.1:
            dec = 4
        elif range < 10:
            dec = 3
        elif range < 100:
            dec = 2
        elif range < 10000:
            dec = 1
        else: 
            dec = 0
        return f"{coord:#.{dec}f}"


    @override
    def resizeEvent(self, ev):

        # update position reset button 
        if self._resetBtn is not None:                  # item could be already closed 
            btnRect = self.mapRectFromItem(self._resetBtn, self._resetBtn.boundingRect())
            y = self.size().height() - btnRect.height()
            self._resetBtn.setPos(20, y+3)              # right aside autoBtn

        super().resizeEvent (ev)


    @property 
    def has_reset_button (self) -> bool:
        """ reset view button in the lower left corner"""
        # to be overridden
        return True 
    

    @override
    def updateButtons(self):
        # PlotItem override to show/hide reset button
        super().updateButtons ()
        try:
            if self.mouseHovering and not self.buttonsHidden and self._vb_state_changed: #  and not all(self.vb.autoRangeEnabled()):
                if self._resetBtn is not None: 
                    self._resetBtn.show()
                self._coordItem.show()
            else:
                if self._resetBtn is not None: 
                    self._resetBtn.hide()
                self._coordItem.hide()
        except RuntimeError:
            pass  # this can happen if the plot has been deleted.


    def _resetBtn_clicked(self):
        """ reset button was clicked - set initial view range"""

        if self._vb_state_changed:
            self.setup_viewRange ()
            self._vb_state_changed = False


    def _reset_prepare (self):
        """ prepare reset button - save initial range"""
        self.sigRangeChanged.connect(self._viewRangeChanged)


    def _viewRangeChanged (self): 
        """ slot - view Range changed"""
        self._vb_state_changed = True 


    @property
    def viewBox (self) -> pg.ViewBox:
        """ viewBox of self""" 
        return self.getViewBox()

    def data_object (self): 
        # to be ooverloaded - or implemented with semantic name 
        if callable(self._getter):
            return self._getter()
        else: 
            return self._getter


    def data_list (self): 
        # to be overloaded - or implemented with semantic name        
        if isinstance (self.data_object(), list):
            return self.data_object()
        else: 
            return [self.data_object()]   


    def refresh(self): 
        """ refresh my artits and section panel """

        refresh_done = False

        if self._section_panel is not None: 
            # refresh artists only if self section is switched on 
            if self.section_panel.switched_on:
                self.refresh_artists()          # first artist and then panel 
                refresh_done = True

            self.section_panel.refresh()

        else: 
            self.refresh_artists() 
            refresh_done = True

        # initial setup of viewRange if artists were shown by refresh

        if not self._viewRange_set and refresh_done:
            self.setup_viewRange()   
            self._viewRange_set = True

        # plot title and sub title with default values of class

        self.plot_title () 


    def setup_artists (self):
        """ create and setup the artists of self"""
        # must be implemented by subclass
        
        logger.warning (f"{self} no artists are setup")


    def setup_viewRange (self):
        """ define view range of this plotItem"""
        # must be implemented by subclass
        
        logger.warning (f"{self} no view range defined")


    def setup_axis (self):
        """ setup axis of of this plotItem"""
        # can be implemented by subclass
        
        # define a constant width 
        yaxis : pg.AxisItem = self.getAxis ('left')
        yaxis.setWidth (40)


    def refresh_artists (self):
        """ refresh all artists of self - can be overridden"""
        for artist in self._artists:
            artist.refresh()


    def plot_title (self, 
                    title : str|None = None,
                    title_size : int = None,
                    title_color : str|QColor = None,
                    subtitle : str|None = None, 
                    align :str ='left', 
                    offset : tuple = (50,5)):
        """plot a title, optionally with a sub title, at fixed position

        Args:
            title: optional - default is class.title
            title_size: optional - title font size in pt 
            title_color: optional - title color as String or QColor
            subtitle: optional - default is class.subtitle
            align: aligned to the left or right of th item .
            offset: from the upper left (right) corner in pixel  
        """

        if title is None:
            if not self.title: return                       # show no title  
            title = self.title

        if title_size is None:
            title_size = Artist.SIZE_HEADER  

        if subtitle is None:
            subtitle = self.subtitle

        # remove existing title item 
        if isinstance (self._title_item, pg.LabelItem):
            if self._title_item:
                self.scene().removeItem (self._title_item)      # was added directly to the scene via setParentItem
            if self._subtitle_item:
                self.scene().removeItem (self._subtitle_item)    
       

        if align == 'left':
            parentPos = (0.0)                               # parent x starts at PlotItem (including axis)       
            itemPos   = (0,0)
        else:  
            parentPos = (0.98,0)
            itemPos   = (1,0)

        if title_color is None: 
            color = Artist.COLOR_HEADER
        else:
            color = title_color

        p1 = pg.LabelItem(title, color=color, size=f"{title_size}pt")    

        p1.setParentItem(self)                            # add to self (Diagram Item) for absolute position 
        p1.anchor(itemPos=itemPos, parentPos=parentPos, offset=offset)
        p1.setZValue(5)
        self._title_item = p1

        if subtitle is not None:
            sub_offset = (offset[0], offset[1]+30)
            p2 = pg.LabelItem(subtitle, color=QColor(Artist.COLOR_LEGEND), size=f"{Artist.SIZE_NORMAL}pt")    
            p2.setParentItem(self)                            # add to self (Diagram Item) for absolute position 
            p2.anchor(itemPos=itemPos, parentPos=parentPos, offset=sub_offset)
            p2.setZValue(5)
            self._subtitle_item = p2 



    @property
    def section_panel (self) -> Edit_Panel | None:
        """ small section panel representing self in view panel"""
        # overload for constructor 
        return  self._section_panel
    



#!/usr/bin/env pythonnowfinished
# -*- coding: utf-8 -*-
"""  

    Part of Optimizer Model to handle Xoptfoil2 result files 

    |-- Case                        - a single optimization case, define specs for an optimization 
        |-- Results                 - proxy of the Xopfoil2 result files 
            |-- Reader_Airfoils     - read airfoil designs of optimzation 
            |-- Reader ...

                                          
"""

import os
import sys
import shutil
from datetime import datetime, timedelta
from pathlib import Path


# let python find the other modules in the dir of self  
sys.path.append(Path(__file__).parent)

from base.common_utils      import * 
from base.spline            import HicksHenne
from model.airfoil          import Airfoil, Airfoil_Bezier, Airfoil_Hicks_Henne, usedAs
from model.airfoil          import GEO_BASIC, Line
from model.polar_set        import * 
from model.xo2_driver       import Xoptfoil2
from model.xo2_input        import GEO_OPT_VARS

import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.WARNING)


#-------------------------------------------------------------------------------
# Results - holds the different Result classes 
#-------------------------------------------------------------------------------


class Xo2_Results:

    TIME_REF_FILE   = 'Optimization_History.csv' 

    def __init__(self, workingDir, outName, remove_result_dir = False, remove_airfoil = False ):

        self._outName       = outName                                           # xo2 result airfoil fielName (withput extension) 
        self._workingDir    = workingDir                                        # working dir of optimizer

        resultDir_rel       = outName + Xoptfoil2.RESULT_DIR_POSTFIX            # directory where the designs are generated
        self._resultDir     = os.path.join (workingDir, resultDir_rel)          # directory where the designs are generated

        # optionally remove existing result_dir 

        if remove_result_dir and os.path.isdir (self.resultDir):
            shutil.rmtree(self.resultDir, ignore_errors=True)

        # optionally remove out airfoil (avoid race conditions as xo2 is slow in deleting at a new optimization) 

        if remove_airfoil:
            pathfileName = os.path.join (workingDir, outName + Airfoil.Extension)
            if os.path.isfile (pathfileName): os.remove (pathfileName)
            pathfileName = os.path.join (workingDir, outName + Airfoil_Bezier.Extension)
            if os.path.isfile (pathfileName): os.remove (pathfileName)
            pathfileName = os.path.join (workingDir, outName + Airfoil_Hicks_Henne.Extension)
            if os.path.isfile (pathfileName): os.remove (pathfileName)

        # results reader  

        self._reader_airfoils              = Reader_Airfoils (self.resultDir)        
        self._reader_airfoils_hh           = Reader_Airfoils_HH (self.resultDir)        
        self._reader_airfoils_bezier       = Reader_Airfoils_Bezier (self.resultDir)    
        self._reader_opPoints              = Reader_OpPoints (self.resultDir)            
        self._reader_geoTargets            = Reader_GeoTargets (self.resultDir)          
        self._reader_optimization_history  = Reader_Optimization_History (self.resultDir) 

        # final airfoils 

        self._airfoil_final         = None
        self._airfoil_final_bezier  = None
        self._airfoil_final_hh      = None


    @property
    def workingDir (self): 
        """ working dir of optimizer - absolut"""
        return self._workingDir

    @property
    def resultDir (self): 
        """ directory with optimizer results - absolut"""
        return self._resultDir

    @property 
    def airfoil_final (self) -> Airfoil | None:
        """ the final airfoil as result of optimization - None if not generated """

        if self._airfoil_final is None:
            self._airfoil_final = self._get_airfoil_final (self.workingDir, self._outName, Airfoil.Extension)
        return self._airfoil_final


    @property 
    def airfoil_final_bezier (self) -> Airfoil | None:
        """ the final airfoil Bezier as result of optimization - None if not generated """

        if self._airfoil_final_bezier is None:
            self._airfoil_final_bezier = self._get_airfoil_final (self.workingDir, self._outName, Airfoil_Bezier.Extension)
        return self._airfoil_final_bezier


    @property 
    def airfoil_final_hh (self) -> Airfoil | None:
        """ the final airfoil Hicks Henneas result of optimization - None if not generated """

        if self._airfoil_final_hh is None:
            self._airfoil_final_hh = self._get_airfoil_final (self.workingDir, self._outName, Airfoil_Hicks_Henne.Extension)
        return self._airfoil_final_hh


    def _get_airfoil_final (self, workingDir : str, outName: str, extension : str) -> Airfoil | None:
        """ if exists return final airfoil for extension .dat or .bez or .hh"""

        fileName = outName + extension
        if os.path.isfile (os.path.join (workingDir, fileName)):
            airfoil =  Airfoil.onFileType (fileName, workingDir=workingDir)
            airfoil.load()
            airfoil.set_usedAs (usedAs.FINAL)
            return airfoil 



    @property
    def nSteps (self) -> int:
        """ number of optimization steps imported (up to now) - "0" excluded""" 
        nSteps = len(self._reader_optimization_history.steps)
        return nSteps if nSteps == 0 else nSteps - 1


    @property
    def nDesigns (self) -> int:
        """ number of Designs steps imported (up to now) - "0" excluded""" 
        nDesigns = 0 
        for step in self._reader_optimization_history.steps [::-1]:
            if step.design != -1: 
                nDesigns = step.design
                break
        return nDesigns 


    @property 
    def improvement (self):
        """ improvement in fraction of 1.0 reached up to now - else 0.0 """
        steps = self._reader_optimization_history.steps
        if steps: 
            improvement = steps[-1].improvement / 100.0                     # xoptfoil2 returns % 
        else: 
            improvement = 0.0 
        return improvement


    @property
    def hasProbablyResults (self) -> bool:
        """ does result dir exist? or isRunning? Then there *should* be results"""

        hasResults = False
        if os.path.isdir(self.resultDir):
            hasResults = True 
        return hasResults


    @property 
    def isFinished (self) -> bool:
        """ is optimization finished - performance summary will exist"""

        return self.hasProbablyResults and os.path.isfile (os.path.join (self.resultDir, self.TIME_REF_FILE)) 


    def time_elapsed (self) -> str:
        """ returns hours, minutes, seconds how long self was running as string hh:mm:ss"""

        last_write_dt  = self._get_dateTime_last_write() 
        first_write_dt = self._get_dateTime_first_write()

        if last_write_dt and first_write_dt:
            delta = last_write_dt - first_write_dt
            hours, remainder = divmod(delta.total_seconds(), 3600)
            minutes, seconds = divmod(remainder, 60)
        else:
            hours, minutes, seconds = 0, 0, 0 

        if hours > 0:
            return f"{int(hours)}:{int(minutes)}:{int(seconds):02d}"
        else: 
            return f"{int(minutes)}:{int(seconds):02d}"

    @property
    def date_time_last_write (self) -> datetime:
        """ youngest date of written results"""
        return self._get_dateTime_last_write()        

    @property
    def steps (self) -> list ['Optimization_History_Entry']:
        """ optimization steps imported (up to now)""" 
        return self._reader_optimization_history.steps


    @property
    def designs_opPoints (self):        
        """ op point results of the designs """
        return self._reader_opPoints.designs


    @property
    def designs_geoTargets (self):        
        """ geo targets results of the designs """
        return self._reader_geoTargets.designs

    @property
    def designs_airfoil (self):        
        """ airfoil of the designs either HicksHenne or Bezier """

        if self._reader_airfoils_hh.hasResults:
            return self._reader_airfoils_hh.designs
        elif self._reader_airfoils_bezier.hasResults:
            return self._reader_airfoils_bezier.designs
        elif self._reader_airfoils.hasResults:
            return self._reader_airfoils.designs
        else: 
            return []


    # ---- Methods -------------------------------------------


    def set_results_could_be_dirty (self): 
        """ mark results of Reader as dirty so they will be re-read at next access"""

        self._reader_airfoils.set_results_could_be_dirty(True)
        self._reader_airfoils_hh.set_results_could_be_dirty(True)
        self._reader_airfoils_bezier.set_results_could_be_dirty(True)
        self._reader_opPoints.set_results_could_be_dirty(True)
        self._reader_geoTargets.set_results_could_be_dirty(True)
        self._reader_optimization_history.set_results_could_be_dirty(True)


    def remove_resultDir (self):
        """ 
        removes result directory - handle with care ...
        """ 

        shutil.rmtree(self.resultDir, ignore_errors=True)


    def _get_dateTime_last_write (self) -> datetime:
        """" dateTime of reference file in result dir if it exists - else None"""
        ref_file = os.path.join (self.resultDir, self.TIME_REF_FILE)  
        if os.path.isfile (ref_file):    
            ts = os.path.getmtime(ref_file)                 # file modification timestamp of a file
            dt = datetime.fromtimestamp(ts)                 # convert timestamp into DateTime object
        else:
            dt = None
        return dt


    def _get_dateTime_first_write (self) -> datetime:
        """" dateTime of file first written which is Xo2 seed airfoil (normalized)"""

        files = os.listdir(self.resultDir) if os.path.isdir (self.resultDir) else []
        if files:
            path_files = [os.path.join(self.resultDir, f) for f in files]
            oldest_file = min(path_files, key=os.path.getctime)
            ts = os.path.getmtime (oldest_file)             # file modification timestamp of a file
            dt = datetime.fromtimestamp(ts)                 # convert timestamp into DateTime object
        else: 
            dt = None
        return dt  

#-------------------------------------------------------------------------------
# Optimization History step entry  
#-------------------------------------------------------------------------------


class Optimization_History_Entry:
    """ 
    An entry in the optimization hostory    
    """
    def __init__(self):
        """
        New entry in the optimization history with Xoptfoil2 data 
        """

        self.step           = 0                # step number in optimization 
        self.improvement    = None             # % improvement to prior design 
        self.objective      = 1.0              # value of objective function 
        self.design_radius  = 0.0              # the design radius of the particles 
        self.design         = -1               # number of new design if achieved in this step 



#-------------------------------------------------------------------------------
# Geometry target Result  
#-------------------------------------------------------------------------------


class GeoTarget_Result (Polar_Point):
    """ 
    The optimization result of the Geometry target  
    """
    def __init__(self):
        """
        New geo target result from optimization 
        """

        self.optVar         = None             # either CAMBER or THICKNESS 
        self.value          = 0.0 
        self.deviation      = 0.0              # deviation from target / improvement to seed 
        self.distance       = 0.0              # distance from target  / distance from seed 
        self.weighting      = 1.0              # actual weighting during optimization 

    @property
    def deviation_label (self):
        """ deviation as formatted string"""

        dev = round(abs(self.deviation),1)

        if dev >= 10.0:
            lab = f"{self.deviation:2.0f}"
        elif dev >= 1.0:
            lab = f"{self.deviation:3.1f}"
        elif dev == 0.0:
            lab = "0"
        else: 
            lab = f"{self.deviation:0.1f}"
        return lab

    @property
    def value_label (self):
        """ value as formatted string"""
        return f"{self.value:.4f}"
    
    @property
    def distance_label (self):
        """ value as formatted string"""

        dist = round(abs(self.distance),5)

        if dist >= 100.0:
           return f"{self.distance:.1f}"
        elif dist >= 10.0:
           return f"{self.distance:.2f}"
        elif dist >= 1.0:
           return f"{self.distance:.3f}"
        elif dist > 0.0:
            if self.distance > 0.0:
                return ("%7.5f" % self.distance)[1:]          # remove leading 0 
            else: 
                return "-" + ("%7.5f" % self.distance)[2:]    # replace leading -0 with - 
        else:
            return "0.0"
    
    @property
    def weighting_label (self):
        """ weighting as formatted string"""
        lab = f"{self.weighting:3.1f}"
        return lab




#-------------------------------------------------------------------------------
# OpPoint Result  
#-------------------------------------------------------------------------------


class OpPoint_Result (Polar_Point):
    """ 
    The optimization result of Xoptfoil2 for single op point - inherits from opPoint   
    """
    def __init__(self):
        """
        New operating point result  from optimization 
        """

        self.idesign        = None             # belongs to Design i 
        self.iopPoint       = None             # is opPoint i 

        self.distance       = 0.0              # distance from target  / distance from seed 
        self.deviation      = 0.0              # deviation from target / improvement to seed 
        self.flap           = 0.0              # flap setting vom optimzation with flaps
        self.weighting      = 1.0              # actual weighting during optimization 

    @property
    def deviation_label (self):
        """ deviation as formatted string"""

        dev = round(abs(self.deviation),1)

        if dev >= 10.0:
            lab = f"{self.deviation:2.0f}"
        elif dev >= 1.0:
            lab = f"{self.deviation:3.1f}"
        elif dev == 0.0:
            lab = "0"
        else: 
            lab = f"{self.deviation:0.1f}"
        return lab

    @property
    def weighting_label (self):
        """ weighting as formatted string"""

        lab = f"{self.weighting:3.1f}"
        return lab



#-------------------------------------------------------------------------------
# Result Handler - read the various results vom Xoptfoil2
#-------------------------------------------------------------------------------


class Reader_Abstract:
    """ 
    Abstract superclass    
    """

    filename = None 
    objects_text  = ('','')

    DESIGN_DIR_EXT = "_designs"
    DESIGN_NAME_BASE = "Design"

    @classmethod
    def design_fileName (cls, iDesign : int, extension : str) -> str:
        """ returns fileName of design iDesign"""

        return f"{cls.DESIGN_NAME_BASE}{iDesign:4}{extension}"
    
    # ----------------------------------


    def __init__(self, resultDir):
        """Superclass for the different Result Handlers for reading results

        Arguments:
            resultDir -- directory where designs with 'filename' can be found    
        """

        self._results   = []                    # list of designs red 
        self._resultDir = resultDir             # directory where the designs are generated
        self._resultFile_lastDate = None        # last file modification date   
        self._results_could_be_dirty = False    # flag that _results could be outdated 

        # read and load results the first time 

        self.read_results()


    def __repr__(self) -> str:
        """ nice print string"""
        return f"<{type(self).__name__}>"


    @property
    def resultPathFile (self): 
        """ the file path of the file with designs """
        if self.filename is None:
            raise ValueError ("Filename for design file not set") 
        return os.path.join(self._resultDir, self.filename)

    @property 
    def results (self): 
        """ list of (abstract) results - could be designs or steps"""

        # re-read results if dirty flag set 
        if self._results_could_be_dirty: 

            n = self.read_results ()

            logger.debug (f"read {n} results after dirty in {self.__class__.__name__}")
            self._results_could_be_dirty = False
        return self._results 

    
    @property 
    def designs (self): 
        """ List of designs of type Airfoil or Polar or ..."""
        return self.results

    @property 
    def hasResults (self) -> bool: 
        """ are there results? will read if dirty """
        return len(self.results) > 0 

    @property 
    def nResults (self) -> int: 
        """ number of result objects already red in"""
        return len(self._results)


    def set_results_could_be_dirty (self, aBool: bool):
        """ set the dirty flag for self.results - will be re-read when accessed next time"""
    
        self._results_could_be_dirty = aBool


    def read_results (self) -> int:
        """ Reads new design, create objects and add them to 'designs' 

        Returns:
            n_new -- number of new results added
        """    

        n_new = 0                                           # n new results red

        # read only if file changed since last time 
        if self._is_younger_than_last_read (self._resultFile_lastDate, self.resultPathFile):

            from timeit import default_timer as timer
            start = timer()
            # ...
            # print("Time ", timer() - start)  

            try: 
                f = open(self.resultPathFile, 'r')
            except:
                logger.error (f"Couldn't read '{self.resultPathFile}' to get designs")
                return n_new

            # read complete file 
                      
            file_lines = f.readlines()
            f.close()

            time_read = timer() - start 

            # save current file date

            ts = os.path.getmtime(self.resultPathFile)                       # file modification timestamp of a file
            self._resultFile_lastDate = datetime.fromtimestamp(ts)      # convert timestamp into DateTime object

            n_before = self.nResults

            # parse line, create objects, add to result list 
            start = timer()

            n_new = self._load_results(file_lines)          # overloaded in sub classes

            time_load = timer() - start 

            # nice message print 
            if n_new > 0: 
                if n_new == 1:
                    object_name = self.objects_text[0]
                else: 
                    object_name = self.objects_text[1]
                if n_before == 0: 
                    new_text = ''
                else:
                    new_text = 'new '
                if time_load < 0.01 and time_read < 0.005:
                    logger.debug (f"{self} imported {n_new} {new_text}{object_name}  (Time read: {time_read:.4f}s, load: {time_load:.4f}s)")
                elif n_new == 1: 
                    logger.warning (f"{self} importing {len(file_lines)} lines takes too long (Time read: {time_read:.4f}s, load: {time_load:.4f}s)")
            else: 
                logger.warning (f"{self} nothing to import")

        return n_new


    def _load_results (self, file_lines): 
        """ parse new lines and create design results """
        return 0                                           # must be over loaded 


    def _is_younger_than_last_read (self, lastDate, aResultFile):
        """ checks if the current file is younger than the version at last read"""

        if  aResultFile is None or not os.path.isfile (aResultFile):
            return False                                # no valid file to check
        
        if lastDate is None:
            return True                                 # first time - no last date available

        ts = os.path.getmtime(aResultFile)              # file modification timestamp of a file
        current = datetime.fromtimestamp(ts)            # convert timestamp into DateTime object

        if lastDate != current:
            return True
        else:
            return False                                # file didn't change

 

class Reader_Optimization_History (Reader_Abstract):
    """
    The Xoptfoil2 optimization history with steps, objective function, design radius 
    """

    filename = 'Optimization_History.csv'
    objects_text  = ('optimization step', 'optimization steps')

    @property
    def steps (self) -> list [Optimization_History_Entry]:
        """ no of steps - will re-read if dirty!"""
        return self.results 


    def _load_results (self, file_lines):
        """ Parse file_lines and create new history objects  

        Arguments:
            file_lines -- (new) lines of the file 
        """
        # File format 
        #
        #   Iter;Design;  Objective;  % Improve; Design rad
        #      0;      ;  1.0000000;  0.0000000;  0.1459420
        #      1;     1;  0.9729227;  2.7077310;  0.1433520

        n_new = 0 
        for i, line in enumerate(file_lines):

            vals = [val.strip() for val in line.split(';')]             # will remove all extra spaces

            if vals[0] == 'Iter' and vals[1] == 'Design':               # Header 
                pass
            else:
                istep = int (vals[0])
                if vals[1] != '':
                    idesign = int(vals[1])                              # step with new design
                else: 
                    idesign = -1                                        # step with no new design                    
                data =  [float(i) for i in vals[2:]]               
                n_new += self.add_history_entry (istep, idesign, data)
        return n_new


    def add_history_entry (self, istep, idesign, data):
        """ add a Optimization_Step to steps  """

        n_new = 0 
        if  istep == self.nResults:             # steps start with istep=0

            history_entry = Optimization_History_Entry()

            history_entry.objective     = data[0]
            history_entry.improvement   = data[1]
            history_entry.design_radius = data[2]
            history_entry.step          = istep
            history_entry.design        = idesign
            
            self._results.append(history_entry) 
            n_new += 1           

        elif istep < self.nResults:              # we have it already 
            pass                                     
        else:                                       #  there would be a gap in design list
            raise ValueError ("Add new history step: Index %i doesn't fit" %istep)
        return n_new
        



class Reader_OpPoints (Reader_Abstract):
    """
    The Xoptfoil2 results for the operating points during an optimization 
    """

    filename = 'Design_OpPoints.csv'
    objects_text  = ('set of op results', 'sets of op results')

    def _load_results (self, file_lines):
        """ Parse file_lines and create new design objects lines of the file freshly red into 

        Arguments:
            file_lines -- (new) lines of the file 
        """
        # File format 
        #
        #    No; iOp;      alpha;         cl;         cd;         cm;       xtrt;       xtrb;       dist;        dev;     flap;   weight
        #     0;   1;  -3.713445;  -0.250000;   0.012932;  -0.037740;   0.982203;   0.026582;  -0.001832; -16.507213;   0.0000;     1.0
        #     0;   2;  -2.236515;  -0.050000;   0.008851;  -0.046911;   0.956186;   0.533244;  -0.001791; -25.370629;   0.0000;     0.5

        n_new = 0 
        idesign_old = len (self._results) 
        ops_results = []

        for i, line in enumerate(file_lines):

            vals = [val.strip() for val in line.split(';')]             # will remove all extra spaces

            if vals[0] == 'No' and vals[1] == 'iOp':   # Header 
                pass
            else:
                idesign = int (vals[0])

                if idesign == idesign_old: 
                    # same design group - add sub object to list 
                    ops_results.append ([float(i) for i in vals[2:]])   # convert list to float

                if idesign > idesign_old :  
                    # new design group - flush old one 
                    n_new += self.add_opPoints_result (idesign_old, ops_results)
                    # start new one with current sub object
                    idesign_old = idesign 
                    ops_results = [[float(i) for i in vals[2:]]]   

                if i == (len(file_lines)-1):
                    # last record - flush current design group 
                    n_new += self.add_opPoints_result (idesign, ops_results)               
      
        return n_new


    def add_opPoints_result (self, idesign, ops_results_list):
        """ add array of OpPoint_Result to designs """

        n_new = 0 
        if idesign == len (self._results):          # new, next design in list 

            ops = []
            for iop, op_result_list in enumerate (ops_results_list): 

                if len(op_result_list) >= 10: 
                    op_result = OpPoint_Result ()

                    op_result.idesign   = idesign 
                    op_result.iopPoint  = iop 

                    op_result.alpha     = op_result_list[0]
                    op_result.cl        = op_result_list[1]
                    op_result.cd        = op_result_list[2]
                    op_result.cm        = op_result_list[3]
                    op_result.xtrt      = op_result_list[4]
                    op_result.xtrb      = op_result_list[5]
                    op_result.distance  = op_result_list[6]
                    op_result.deviation = op_result_list[7]
                    op_result.flap      = op_result_list[8]
                    op_result.weighting = op_result_list[9]
                    ops.append(op_result)
                else: 
                    logger.error (f"Format of '{self.filename}' doesn't fit. Skipping op point data...")       

            self._results.append(ops)   
            n_new += 1  

        elif idesign < len (self._results):         # we have it already 
            pass                                     
        elif len (self._results) == 0:              # we are already in error mode
            pass
        else:                                       #  there would be a gap in design list
            raise ValueError ("Add new design: Index %i doesn't fit" %idesign)
        return n_new



class Reader_GeoTargets (Reader_Abstract):
    """
    The results for the geometry targets during an optimization 
    """

    filename = 'Design_GeoTargets.csv'
    objects_text  = ('set of geo target results', 'sets of geo target results')

    def _load_results (self, file_lines):
        """ Parse file_lines and create new design objects lines of the file freshly red into 

        Arguments:
            file_lines -- (new) lines of the file 
        """
        # File format 
        #
        #    No; iGeo;        type;       val;        dev;   weight 
        #     0;    1;    'Camber';   0.01711; -16.507213;      1.0   
        #     0;    2; 'Thickness';   0.08201; -16.507213;      1.0     

        n_new = 0 
        idesign_old = 0 
        geo_results = []

        for i, line in enumerate(file_lines):

            vals = [val.strip() for val in line.split(';')]             # will remove all extra spaces

            if vals[0] == 'No' and vals[1] == 'iGeo':   # Header 
                pass
            else:
                idesign = int (vals[0])

                if idesign == idesign_old: 
                    # same design group - add sub object to list 
                    geo_results.append ([vals[2]] + [float(i) for i in vals[3:]]) 

                if idesign != idesign_old :  
                    # new design group - flush old one 
                    n_new += self.add_geoTarget_result (idesign_old, geo_results)
                    # start new one with current sub object
                    idesign_old = idesign 
                    geo_results = [[vals[2]] + [float(i) for i in vals[3:]]]

                if i == (len(file_lines)-1):
                    # last record - flush current design group 
                    n_new += self.add_geoTarget_result (idesign, geo_results)               

        return n_new


    def add_geoTarget_result (self, idesign, geos_results_list):
        """ add array of GeoTarget_Result to designs """

        n_new = 0 
        if idesign == len (self._results):          # new, next design in list 

            geos = []
            for geo_result_list in geos_results_list: 

                if len(geo_result_list) >= 2: 
                    geo_result = GeoTarget_Result ()

                    optVar_result : str = geo_result_list[0]
                    for optVar in GEO_OPT_VARS:                         # support "thickness" and "Thickness"
                        if optVar_result.lower() == optVar.lower():
                            geo_result.optVar    = optVar
                            geo_result.value     = geo_result_list[1]
                            geo_result.distance  = geo_result_list[2]
                            geo_result.deviation = geo_result_list[3]
                            geo_result.weighting = geo_result_list[4]
                            geos.append(geo_result)
                else: 
                    logger.error (f"Format of '{self.filename}' doesn't fit. Skipping op point data...")       

            self._results.append(geos)  
            n_new += 1  

        elif idesign < len (self._results):         # we have it already 
            pass                                     
        elif len (self._results) == 0:              # we are already in error mode
            pass
        else:                                       #  there would be a gap in design list
            raise ValueError ("Add new design: Index %i doesn't fit" %idesign)
        return n_new




class Reader_Airfoils (Reader_Abstract):
    """
    The airfoils generated during an optimization 
    """
    filename = 'Design_Coordinates.csv'
    objects_text  =('airfoil', 'airfoils')

    def _load_results (self, file_lines : list[str]):
        """ Parse file_lines and create new design objects lines of the file freshly red into 

        Arguments:
            file_lines -- (new) lines of the file 
        """
        # File format 
        #
        #    No;           Name; Coord;         1;         2;         3;     
        #     0;JX-Seed-Reflexed;    x; 1.0000000; 0.9905321; 0.9797401; 0.96
        #     0;JX-Seed-Reflexed;    y; 0.0001498; 0.0003774; 0.0007564; 0.00

        x,y = [], []
        n_new = 0 
        for i, line in enumerate(file_lines):

            vals = [val.strip() for val in line.split(';')]             # will remove all extra spaces

            if vals[0] == 'No' and vals[1] == 'Name':                   # Header 
                pass

            elif vals[2] == 'x' or vals[2] =='y':                       # a valid line 
                idesign = int (vals[0])

                name    = vals[1]
                coord   = vals[2]
                if coord == 'x':
                    x = [float(i) for i in vals[3:]] 
                else: 
                    y = [float(i) for i in vals[3:]] 
                if x and y: 
                    n_new += self.add_airfoil_design (idesign, name, x, y) 
                    x, y = [], []

            else:
                logger.error ("Invalid coordinates file format for designs - skipped.")
                break 
        return n_new


    def add_airfoil_design (self, idesign, name, x, y):
        """ add a new airfoil design to my designs """

        n_new = 0 
        if idesign == len (self._results):          # new, next design in list 

            # create airfoil - set its file path to resultDir for lazy save to generate polar
            #                  use basic Geometry (not splined) for faster evaluation
            fileName = self.design_fileName (idesign, Airfoil.Extension )

            airfoil = Airfoil (name=name, workingDir=self._resultDir, geometry=GEO_BASIC)

            airfoil.set_xy (x,y)
            airfoil.set_pathFileName (fileName, noCheck=True)               # no check - it doesn't exist
            airfoil.set_usedAs (usedAs.DESIGN)

            # if airfoil file not was already created before, set modify for lazy write 
            if os.path.isfile (airfoil.pathFileName):
                airfoil.set_isModified (False)
            else: 
                airfoil.set_isModified (True)             # up to now airfoil file doesn't exist  

            self._results.append(airfoil)

            n_new = 1           

        elif idesign < len (self._results):         # we have it already 
            pass                                     
        else:                                       #  there would be a gap in design list
            raise ValueError ("Add new design airfoil: Index %i doesn't fit" %idesign)

        return n_new



# -----------------------------------------


class Reader_Airfoils_Bezier (Reader_Abstract):
    """
    The airfoils as bezier definitions generated durng an optimization 
    """

    filename = 'Design_Beziers.csv'
    objects_text  =('bezier airfoil', 'bezier airfoils')

    def _load_results (self, file_lines):
        """ Parse file_lines and create new design objects lines of the file freshly red into 

        Arguments:
            file_lines -- (new) lines of the file 
        """
        # File format 
        #
        #    No;           Name; Side;         p1x;         p1y;         p2x;         p2y;         p3x;         p3y;         
        #     0;JX-Seed-Reflexed_bezier;  Top;  0.00000000;  0.00000000;  0.00000000;  0.02508150;  0.12899340;  0.10043424
        #     0;JX-Seed-Reflexed_bezier;  Bot;  0.00000000;  0.00000000;  0.00000000; -0.01452336;  0.09103726; -0.03561080       #    No; Side;         p1x;         p1y;         p2x;         p2y;     

        pxy_top, pxy_bot = [], []
        n_new = 0 

        for i, line in enumerate(file_lines):

            vals = [val.strip() for val in line.split(';')]             # will remove all extra spaces

            if vals[0] == 'No' and vals[1] == 'Name':   # Header 
                pass

            elif vals[2] == 'Top' or vals[2] =='Bot':  # a valid line 
                idesign = int (vals[0])
                side = vals[2]
                if side == 'Top':
                    pxy_top = [float(i) for i in vals[3:]] 
                else: 
                    pxy_bot = [float(i) for i in vals[3:]] 
                if pxy_top and pxy_bot: 
                    n_new += self.add_airfoil_design (idesign, vals[1], pxy_top, pxy_bot) 
                    pxy_top, pxy_bot = [], []
            else:
                logger.error ("Invalid Bezier file format for designs - skipped.")
                break 
        return n_new


    def add_airfoil_design (self, idesign, name, pxy_top, pxy_bot):
        """ add a new bezier based airfoil design to my designs """

        n_new = 0 
        if idesign == len (self._results):          # new, next design in list 

            # create bezier airfoil out of bezier upper and lower 
            #  - set its file path to resultDir for lazy save to generate polar
            airfoil = Airfoil_Bezier (name=name, workingDir=self._resultDir, )

            fileName = self.design_fileName (idesign, Airfoil_Bezier.Extension )

            airfoil.set_pathFileName (fileName, noCheck=True)

            px, py = [], []
            for i, x_or_y in enumerate(pxy_top):
                if i % 2 == 0: px.append(x_or_y)
                else:          py.append(x_or_y)
            airfoil.set_newSide_for (Line.Type.UPPER, px, py)

            px, py = [], []
            for i, x_or_y in enumerate(pxy_bot):
                if i % 2 == 0: px.append(x_or_y)
                else:          py.append(x_or_y)
            airfoil.set_newSide_for (Line.Type.LOWER, px, py)

            airfoil.set_usedAs (usedAs.DESIGN)
            airfoil.set_isLoaded (True)
            airfoil.set_isModified (True)             # up to now airfoil file doesn't exist  

            self._results.append(airfoil)

            n_new = 1           

        elif idesign < len (self._results):         # we have it already 
            pass                                     
        else:                                       #  there would be a gap in design list
            raise ValueError ("Add new design airfoil: Index %i doesn't fit" %idesign)

        return n_new




class Reader_Airfoils_HH (Reader_Abstract):
    """
    The airfoils as Hicks Henne definitions generated during an optimization 
    """

    filename = 'Design_Hicks.csv'
    objects_text  = ('hh airfoil', 'hh airfoils')

    def __init__(self, *args, **kwargs):

        self._seed_x = None                             # x,y coordinates of seed 
        self._seed_y = None 
        self._seed_name = None

        super().__init__(*args,  **kwargs)


    def _load_results (self, file_lines):
        """ Parse file_lines and create new design objects lines of the file freshly red into 
        """

        # File format 
        #
        #    No;           Name; Coord;         1;         2;         3;         4;         5;         6;         
        #     0;X2-Seed-Rearload-repan-preset;    x; 1.0000000; 0.9930137; 0.9830685; 0.9688287; 0.9539262; 0.9386
        #     0;X2-Seed-Rearload-repan-preset;    y; 0.0001545; 0.0010889; 0.0024129; 0.0042939; 0.0062408; 0.0082
        #    No;           Name; Side;     hh1_str;     hh1_loc;     hh1_wid;     hh2_str;     hh2_loc;     hh2_wi
        #     1;JX-GT-10v3~1;  Top;  0.00026016;  0.19854563;  0.99892745;  0.00044837;  0.40176901;  0.99883876; 
        #     1;JX-GT-10v3~1;  Bot; -0.00018333;  0.24907524;  1.00275824;  0.00036286;  0.49963857;  1.00189794; 

        hhs_top, hhs_bot = [], []
        n_new = 0 

        for i, line in enumerate(file_lines):

            if i == 0 or i == 3:                                        # Header 
                continue
            if i == 1 and self._seed_x :                                # seed coordinates already read
                continue 
            if i == 2 and self._seed_y :                                # seed coordinates already read
                continue 

            vals = [val.strip() for val in line.split(';')]             # will remove all extra spaces

            if i == 1: 
                self._seed_x = [float(i) for i in vals[3:]] 
                self._seed_name = vals[1]
            elif i == 2: 
                self._seed_y = [float(i) for i in vals[3:]] 
                n_new += self.add_airfoil_design (0)                    # add design 0 = seed airfoil

            elif vals[2] == 'Top' or vals[2] =='Bot':                   # a valid line with hhs
                idesign, name, side = int (vals[0]), vals[1], vals[2]
                if side == 'Top':
                    hhs_top = [float(i) for i in vals[3:]] 
                else: 
                    hhs_bot = [float(i) for i in vals[3:]] 
                if hhs_top and hhs_bot: 
                    n_new += self.add_airfoil_design (idesign, name, hhs_top, hhs_bot) 
                    hhs_top, hhs_bot = [], []
            else:
                logger.error ("Invalid Hicks Henne file format for designs - skipped.")
                break 
        return n_new


    def add_airfoil_design (self, idesign, name=None, top_hh_vals=None, bot_hh_vals=None):
        """ add a new Hicks Henne based airfoil design to my designs """

        n_new = 0 

        #  design #0 is the seed airfoil 

        if idesign == 0  and len (self._results) == 0: 

            airfoil = Airfoil (name=self._seed_name, geometry=GEO_BASIC)
            airfoil.set_xy (self._seed_x,self._seed_y)
            airfoil.set_pathFileName (os.path.join(self._resultDir, self._seed_name + '.dat'), noCheck=True)
            airfoil.set_isModified (True)             # up to now airfoil file doesn't exist  
            airfoil.set_usedAs (usedAs.DESIGN)
            self._results.append(airfoil)

        # create hicks henne airfoil out of upper and lower hh functions 
        #  - set its file path to resultDir for lazy save to generate polar

        elif idesign == len (self._results):          # new, next design in list 

            top_hhs = self._get_hhs (top_hh_vals)
            bot_hhs = self._get_hhs (bot_hh_vals)

            airfoil = Airfoil_Hicks_Henne (name=name, workingDir=self._resultDir)

            fileName = self.design_fileName (idesign, Airfoil_Hicks_Henne.Extension )
            airfoil.set_pathFileName (fileName, noCheck=True)
            airfoil.set_hh_data (name, self._seed_name, self._seed_x, self._seed_y, top_hhs, bot_hhs)
            airfoil.set_usedAs (usedAs.DESIGN)
            airfoil.set_isModified (True)             # up to now airfoil file doesn't exist  
            self._results.append(airfoil)
            n_new += 1           

        elif idesign < len (self._results):         # we have it already 
            pass                                     
        else:                                       #  there would be a gap in design list
            raise ValueError ("Add new design airfoil: Index %i doesn't fit" %idesign)

        return n_new


    def _get_hhs (self, hh_vals): 
        """ extract hh function out of array of hh values """

        if len(hh_vals)%3 != 0: 
            raise ValueError ("no valid Hicks Henne data array")

        hhs = []
        nhh = int(len(hh_vals) / 3)
        for ihh in range(nhh):
                strength = hh_vals[ihh*3] 
                location = hh_vals[ihh*3 + 1] 
                width    = hh_vals[ihh*3 + 2] 
                hhs.append (HicksHenne (strength, location, width ))
        return hhs



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""  

UI panels 

"""

import logging

from PyQt6.QtWidgets        import QMenu
from PyQt6.QtGui            import QDesktopServices
from PyQt6.QtCore           import QUrl

from base.widgets           import * 
from base.panels            import Edit_Panel

from model.airfoil          import Airfoil
from model.airfoil_geometry import Geometry, Geometry_Bezier, Curvature_Abstract
from model.airfoil_geometry import Line, Side_Airfoil_Bezier
from model.case             import Case_Abstract, Case_Direct_Design
from model.xo2_driver       import Xoptfoil2

from airfoil_widgets        import * 
from airfoil_dialogs        import Match_Bezier_Dialog, Matcher


logger = logging.getLogger(__name__)
# logger.setLevel(logging.WARNING)


class Panel_Airfoil_Abstract (Edit_Panel):
    """ 
    Abstract superclass for Edit/View-Panels of AirfoilEditor
        - has semantics of App
        - connect / handle signals 
    """

    from app  import Main

    @property
    def app (self) -> Main:
        return self._app 

    @property
    def airfoil (self) -> Airfoil: 
        return self.dataObject

    @property
    def geo (self) -> Geometry:
        return self.airfoil.geo

    @property    
    def case (self) -> Case_Abstract:
        return self.app.case


    @override
    def _set_panel_layout (self ):
        """ Set layout of self._panel """
        # overridden to connect to widgets changed signal

        super()._set_panel_layout ()
        for w in self.widgets:
            w.sig_changed.connect (self._on_airfoil_widget_changed)
        for w in self.header_widgets:
            w.sig_changed.connect (self._on_airfoil_widget_changed)


    def _on_airfoil_widget_changed (self, widget):
        """ user changed data in widget"""
        logger.debug (f"{self} {widget} widget changed slot")
        self.app._on_airfoil_changed ()


    @property
    def mode_modify (self) -> bool:
        """ panel in mode_modify or disabled ? - from App """
        return self.app.mode_modify


    @property
    def mode_optimize (self) -> bool:
        """ panel in mode_modify or disabled ? - from App """
        return self.app.mode_optimize


    @property
    def mode_bezier (self) -> bool:
        """ True if self is in mode_modify and geo is Bezier """
        return self.airfoil.isBezierBased


    @override
    @property
    def _isDisabled (self) -> bool:
        """ overloaded: only enabled in edit mode of App """
        return not self.mode_modify or self.airfoil.isFlapped
    


class Panel_File_View (Panel_Airfoil_Abstract):
    """ File panel with open / save / ... """

    name = 'View Mode'


    @override
    @property
    def shouldBe_visible (self) -> bool:
        """ overloaded: only visible if mode_modify """
        return not (self.mode_modify or self.mode_optimize)

    @property
    def _isDisabled (self) -> bool:
        """ override: always enabled """
        return False
    

    def _on_airfoil_widget_changed (self, *_ ):
        """ user changed data in widget"""
        # overloaded - do not react on self widget changes 
        pass


    def _init_layout (self): 

        l = QGridLayout()
        r,c = 0, 0 
        Airfoil_Select_Open_Widget (l,r,c, colSpan=4, signal=False, 
                                    textOpen="&Open", widthOpen=90, 
                                    get=lambda: self.airfoil, set=self.app.set_airfoil)
        r += 1
        Button (l,r,c, text="&Modify", width=90, 
                set=self.app.modify_airfoil, toolTip="Modify geometry, Normalize, Repanel, Set Flap",
                button_style=button_style.PRIMARY)
        MenuButton (l,r,c+2, text="More...", width=90, 
                menu=self._more_menu(), 
                toolTip="Choose further actions for this airfoil")
        # Button (l,r,c+2, text="&As Bezier", width=90, colSpan=2,
        #         set=self.app.new_as_Bezier, disable=lambda: self.airfoil.isBezierBased,
        #         toolTip="Create new Bezier airfoil based on current airfoil")
        r += 1
        SpaceR (l,r, height=10, stretch=0)
        r += 1
        Button (l,r,c, text="&Optimize...", width=90, 
                set=self.app.optimize_select, 
                toolTip=self._tooltip_optimize,
                disable=lambda: not Xoptfoil2.ready)
        r += 1
        SpaceR (l,r, stretch=1)
        r += 1
        Button (l,r,c, text="&Exit", width=90, set=self.app.close)
        r += 1
        SpaceR (l,r, height=5, stretch=0)        
        l.setColumnStretch (2,2)
        l.setColumnMinimumWidth (1,8)

        return l 
 

    def _more_menu (self) -> QMenu:
        """ create and return sub menu for 'more' actions"""

        menue = QMenu ()

        menue.addAction (MenuAction ("As Bezier based", self, set=self.app.new_as_Bezier, 
                                     disable=lambda: self.airfoil.isBezierBased,
                                     toolTip="Create new Bezier based airfoil of current airfoil"))
        menue.addSeparator ()
        menue.addAction (MenuAction ("Save as...", self, set=self.app.do_save_as,
                                     toolTip="Create a copy of the current airfoil with new name and filename"))
        menue.addAction (MenuAction ("Rename...", self, set=self.app.do_rename,
                                     toolTip="Rename name and/or filename of current airfoil"))
        menue.addAction (MenuAction ("Delete", self, set=self.app.do_delete,
                                     toolTip="Delete current airfoil including all temporary files created by the AirfoilEditor"))
        menue.addAction (MenuAction ("Delete temp files", self, set=self.app.do_delete_temp_files,
                                     toolTip="Delete all temporary files created by the AirfoilEditor just to have a clean directoy again"))
        menue.addSeparator ()
        menue.addAction (MenuAction ("Readme on Github", self, set=self._open_AE_url,
                                     toolTip="Open the Github README file of the AirfoilEditor in a browser"))
        menue.addAction (MenuAction ("Releases on Github", self, set= self._open_releases_url,
                                     toolTip="Open the Github page with the actual release of the AirfoilEditor"))

        menue.setToolTipsVisible(True)

        return menue


    def _open_releases_url (self):
        """ open Github versions in Browser"""

        link = "https://github.com/jxjo/AirfoilEditor/releases"
        QDesktopServices.openUrl(QUrl(link))


    def _open_AE_url (self):
        """ open Github AirfoilEditor repo in Browser"""

        link = "https://github.com/jxjo/AirfoilEditor"
        QDesktopServices.openUrl(QUrl(link))


    def _tooltip_optimize (self) -> str:
        """ dynamic tooltip to handle xo2 not ready"""

        if Xoptfoil2.ready:
            return f"Optimize an airfoil with {Xoptfoil2.NAME}"
        else: 
            return f"Optimization not possible <br><br><b>{Xoptfoil2.NAME} not ready</b><br><br>{Xoptfoil2.ready_msg}"



class Panel_File_Modify (Panel_Airfoil_Abstract):
    """ File panel with open / save / ... """

    name = 'Modifiy Mode'

    @override
    def title_text (self) -> str: 
        """ returns text of title - default self.name"""
        if self.app.airfoil and self.app.airfoil.geo.isBezier:
             return 'Bezier Mode'
        else: 
            return self.name
    
    @property
    def _isDisabled (self) -> bool:
        """ override: always enabled """
        return False
    
    @property
    def airfoilg_org (self) -> Airfoil:
        return self.app.airfoil_org

    @property
    def case (self) -> Case_Direct_Design:
        return self.app.case


    def _init_layout (self): 

        self.set_background_color (**mode_color.MODIFY)

        l = QGridLayout()
        r,c = 0, 0 
        Field (l,r,c, colSpan=3, width=190, get=lambda: self.airfoilg_org.fileName)
        r += 1
        ComboSpinBox (l,r,c, colSpan=2, width=160, get=self.airfoil_fileName, 
                             set=self.set_airfoil_by_fileName,
                             options=self.airfoil_fileNames,
                             signal=False)
        ToolButton   (l,r,c+2, icon=Icon.DELETE, set=self.remove_current_airfoil,
                      hide=lambda: self.case.get_i_from_design (self.airfoil) == 0) # hide Design 0 
        r += 1
        SpaceR (l,r)
        l.setRowStretch (r,2)
        r += 1
        Button (l,r,c,  text="&Finish ...", width=90, 
                        set=lambda : self.app.mode_modify_finished(ok=True), 
                        toolTip="Save current airfoil, optionally modifiy name and leave edit mode")
        r += 1
        SpaceR (l,r, height=5, stretch=0)
        r += 1
        Button (l,r,c,  text="&Cancel",  width=90, 
                        set=lambda : self.app.mode_modify_finished(ok=False),
                        toolTip="Cancel modifications of airfoil and leave edit mode")
        r += 1
        SpaceR (l,r, height=5, stretch=0)
        l.setColumnStretch (3,2)

        return l


    def airfoil_fileName(self) -> list[str]:
        """ fileName of current airfoil without extension"""
        return os.path.splitext(self.airfoil.fileName)[0]


    def airfoil_fileNames(self) -> list[str]:
        """ list of design airfoil fileNames without extension"""

        fileNames = []
        for airfoil in self.case.airfoil_designs:
            fileNames.append (os.path.splitext(airfoil.fileName)[0])
        return fileNames


    def set_airfoil_by_fileName (self, fileName : str):
        """ set new current design airfoil by fileName"""

        airfoil = self.case.get_design_by_name (fileName)
        self.app.set_airfoil (airfoil)


    def remove_current_airfoil (self):
        """ remove current design and set new current design airfoil by fileName"""

        next_airfoil = self.case.remove_design (self.airfoil)

        if next_airfoil: 
            self.app.set_airfoil (next_airfoil)



class Panel_Geometry (Panel_Airfoil_Abstract):
    """ Main geometry data of airfoil"""

    name = 'Geometry'
    _width  = 370

    @override
    def _add_to_header_layout(self, l_head: QHBoxLayout):
        """ add Widgets to header layout"""

        # blend with airfoil - currently Bezier is not supported
        Button (l_head, text="&Blend", width=80,
                set=self.app.do_blend_with, 
                hide=lambda: not self.mode_modify or self.airfoil.isBezierBased,
                toolTip="Blend original airfoil with another airfoil")


    def _init_layout (self): 

        l = QGridLayout()
        r,c = 0, 0 
        FieldF (l,r,c, lab="Thickness", width=75, unit="%", step=0.1,
                obj=lambda: self.geo, prop=Geometry.max_thick,
                disable=lambda: self.airfoil.isBezierBased)
        r += 1
        FieldF (l,r,c, lab="Camber", width=75, unit="%", step=0.1,
                obj=lambda: self.geo, prop=Geometry.max_camb,
                disable=lambda: self.airfoil.isBezierBased or self.airfoil.isSymmetrical)
        r += 1
        FieldF (l,r,c, lab="LE radius", width=75, unit="%", step=0.02,
                obj=lambda: self.geo, prop=Geometry.le_radius,
                disable=lambda: self.airfoil.isBezierBased)
        r += 1
        FieldF (l,r,c, lab="TE gap", width=75, unit="%", step=0.1,
                obj=lambda: self.geo, prop=Geometry.te_gap)
        r += 1
        SpaceR (l,r, height=5)
        r += 1
        Label  (l,r,0,colSpan=4, get=self._messageText, style=style.COMMENT, height=(None,None))

        r,c = 0, 2 
        SpaceC (l,c, stretch=0)
        c += 1 
        FieldF (l,r,c, lab="at", width=75, unit="%", step=0.2,
                obj=lambda: self.geo, prop=Geometry.max_thick_x,
                disable=lambda: self.airfoil.isBezierBased)
        r += 1
        FieldF (l,r,c, lab="at", width=75, unit="%", step=0.2,
                obj=lambda: self.geo, prop=Geometry.max_camb_x,
                disable=lambda: self.airfoil.isBezierBased or self.airfoil.isSymmetrical)
        r += 1
        FieldF (l,r,c, lab="LE curv", width=75, dec=0, disable=True,
                obj=lambda: self.geo.curvature, prop=Curvature_Abstract.max_around_le)
        r += 1
        FieldF (l,r,c, lab="TE curv", width=75, dec=0, disable=True,
                obj=lambda: self.geo.curvature, prop=Curvature_Abstract.max_te,
                style=lambda: style.NORMAL if self.geo.curvature.max_te <2 else style.WARNING)

        l.setColumnMinimumWidth (0,80)
        l.setColumnMinimumWidth (3,60)
        l.setColumnStretch (5,2)
        return l 


    def _messageText (self): 
        """ text to show at bottom of panel"""

        if self.geo.curvature.max_te >= 2:
            text = f"- Curvature at trailing edge is quite high"
        else:
            text = f"Geometry {self.geo.description}"
        return text 



class Panel_Panels (Panel_Airfoil_Abstract):
    """ Panelling information """

    name = 'Panels'
    _width  =  (290, None)

    def _add_to_header_layout(self, l_head: QHBoxLayout):
        """ add Widgets to header layout"""

        # repanel airfoil - currently Bezier is not supported
        Button (l_head, text="&Repanel", width=80,
                set=self.app.do_repanel, hide=lambda: not self.mode_modify,
                disable=lambda: self.geo.isBasic or self.geo.isHicksHenne,
                toolTip="Repanel airfoil with a new number of panels" ) 


    def _init_layout (self):

        l = QGridLayout()

        r,c = 0, 0 
        FieldI (l,r,c, lab="No of panels", disable=True, width=70, style=self._style_panel,
                get=lambda: self.geo.nPanels, )
        r += 1
        FieldF (l,r,c, lab="Angle at LE", width=70, dec=1, unit="°", style=self._style_angle,
                obj=lambda: self.geo, prop=Geometry.panelAngle_le)
        SpaceC (l,c+2, width=10, stretch=0)
        Label  (l,r,c+3,width=70, get=lambda: f"at index {self.geo.iLe}")
        r += 1
        FieldF (l,r,c, lab="Angle min", width=70, dec=1, unit="°",
                get=lambda: self.geo.panelAngle_min[0], )
        r += 1
        SpaceR (l,r,height=5)
        r += 1
        Label  (l,r,0,colSpan=4, get=self._messageText, style=style.COMMENT, height=(None,None))

        l.setColumnMinimumWidth (0,80)
        l.setColumnStretch (c+4,1)
        l.setRowStretch    (r-1,2)
        
        return l
 

    def _on_panelling_finished (self, aSide : Side_Airfoil_Bezier):
        """ slot for panelling (dialog) finished - reset airfoil"""


    def _style_panel (self):
        """ returns style.WARNING if panels not in range"""
        if self.geo.nPanels < 120 or self.geo.nPanels > 260: 
            return style.WARNING
        else: 
            return style.NORMAL


    def _style_angle (self):
        """ returns style.WARNING if panel angle too blunt"""
        if self.geo.panelAngle_le > 175.0: 
            return style.WARNING
        else: 
            return style.NORMAL

    def _messageText (self): 

        text = []
        minAngle, _ = self.geo.panelAngle_min

        if self.geo.panelAngle_le > 175.0: 
            text.append("- Panel angle at LE (%d°) is too blunt" %(self.geo.panelAngle_le))
        if minAngle < 150.0: 
            text.append("- Min. angle of two panels is < 150°")
        if self.geo.panelAngle_le == 180.0: 
            text.append("- Leading edge has 2 points")
        if self.geo.nPanels < 100 or self.geo.nPanels > 200: 
            text.append("- No of panels should be > 100 and < 200")
        
        text = '\n'.join(text)
        return text 




class Panel_Flap (Panel_Airfoil_Abstract):
    """ Flap information and set flap"""

    name = 'Flap'
    _width  = 240

    @override
    @property
    def shouldBe_visible (self) -> bool:
        """ overloaded: only visible if geo is Bezier """
        return (self.mode_modify or self.airfoil.geo.isProbablyFlapped) and not self.mode_bezier


    def _add_to_header_layout(self, l_head: QHBoxLayout):
        """ add Widgets to header layout"""

        # repanel airfoil - currently Bezier is not supported
        Button (l_head, text="Set F&lap", width=80,
                set=self.app.do_flap, hide=lambda: not self.mode_modify,
                disable=self._set_flap_disabled,
                toolTip="Set flap at airfoil" ) 


    def _set_flap_disabled (self) -> bool:
        """ True if set flap is not possible"""

        if self.geo.isBezier or self.geo.isHicksHenne: 
            return True
        elif self.mode_modify:
            return not self.airfoil.flap_setter                       # no flapper, no set flap 
        else:
            return True


    def _init_layout (self):


        geo             = self.airfoil.geo

        l = QGridLayout()
        r,c = 0, 0 

        if self.mode_modify:

            flap_setter         = self.airfoil.flap_setter
            airfoil_flapped = flap_setter.airfoil_flapped if flap_setter else None

            if airfoil_flapped:
                FieldF (l,r,c, lab="Hinge x", width=50, get=lambda: flap_setter.x_flap, dec=1, unit="%")
                r += 1
                FieldF (l,r,c, lab="Flap Angle", width=50, dec=1, unit='°', get=lambda: flap_setter.flap_angle)
                r +=1
                Field (l,r,c, lab="Based on", width=120,
                       get=flap_setter.airfoil_base.fileName)
                r += 1
                SpaceR (l,r, stretch=2)
                r += 1
                lab =Label  (l,r,c, width=None, height=(40,None), colSpan=3, style=style.COMMENT, wordWrap=True,
                        get="As the base airfoil is still available, another flap setting can be applied.")
                lab.setAlignment (ALIGN_BOTTOM)
                l.setRowStretch (r,1)
                l.setColumnMinimumWidth (0,80)
                l.setColumnStretch (2,3)

            elif flap_setter:

                SpaceR (l,r, stretch=2)
                r += 1
                lab =Label  (l,r,c, width=None,  colSpan=3, style=style.COMMENT, wordWrap=True,
                             get="Ready to set flap")
                l.setColumnStretch (2,3)

            elif self.airfoil.isFlapped:
                FieldF (l,r,c, lab="Hinge x", width=50, get=lambda: geo.curvature.flap_kink_at, dec=1, unit="%")
                r += 1
                FieldF (l,r,c, lab="Flap Angle", width=50, dec=1, unit='°', get=lambda: geo.flap_angle_estimated)
                r +=1
                SpaceR (l,r, stretch=2)
                r += 1
                lab =Label  (l,r,c, width=None, colSpan=3, style=style.WARNING,
                             styleRole=QPalette.ColorRole.Window, 
                             get="The airfoil is already flapped.")
                r += 1
                lab =Label  (l,r,c, width=None, height=(40,None), colSpan=3, style=style.COMMENT, wordWrap=True, 
                             get="Choose a Design, which is not flapped to set a flap.")
                lab.setAlignment (ALIGN_BOTTOM)
                l.setRowStretch (r,1)
                l.setColumnMinimumWidth (0,80)
                l.setColumnStretch (2,3)

        else: 
            # mode not modify - info about a flap which is (could be) set
            
            if geo.isFlapped:
                
                FieldF (l,r,c, lab="Hinge x", width=50, get=lambda: geo.curvature.flap_kink_at, dec=1, unit="%")
                r += 1
                FieldF (l,r,c, lab="Flap Angle", width=50, dec=1, unit='°', get=lambda: geo.flap_angle_estimated)
                r +=1
                SpaceR (l,r, stretch=2)
                r += 1
                lab =Label  (l,r,c, width=None, colSpan=3, style=style.COMMENT, wordWrap=True, 
                             get="The airfoil has a set flap.")
                l.setColumnMinimumWidth (0,80)
                l.setColumnStretch (2,3)
                                
            elif geo.isProbablyFlapped:

                SpaceR (l,r, stretch=2)
                r += 1
                lab =Label  (l,r,c, width=None, height=(80,None), colSpan=3, style=style.COMMENT, wordWrap=True, 
                             get="The airfoil is probably flapped, but a kink in the contour couldn't be detected on both sides.")
                lab.setAlignment (ALIGN_BOTTOM)
                l.setColumnStretch (2,3)

        return l

    @override
    def refresh(self, reinit_layout=False):

        # force new layout to show different flpa states 
        return super().refresh(reinit_layout=True)



class Panel_LE_TE  (Panel_Airfoil_Abstract):
    """ info about LE and TE coordinates"""

    name = 'LE, TE'

    _width  = 320

    @override
    @property
    def shouldBe_visible (self) -> bool:
        """ overloaded: only visible if geo is not Bezier """
        return not self.mode_bezier 


    def _add_to_header_layout(self, l_head: QHBoxLayout):
        """ add Widgets to header layout"""

        Button (l_head, text="&Normalize", width=80,
                set=lambda : self.airfoil.normalize(), signal=True, 
                hide=lambda: not self.mode_modify,
                toolTip="Normalize airfoil to get leading edge at 0,0")


    def _init_layout (self): 

        l = QGridLayout()     
        r,c = 0, 0 
        FieldF (l,r,c, lab="Leading edge", get=lambda: self.geo.le[0], width=75, dec=7, style=lambda: self._style (self.geo.le[0], 0.0))
        r += 1
        FieldF (l,r,c, lab=" ... of spline", get=lambda: self.geo.le_real[0], width=75, dec=7, style=self._style_le_real,
                hide=lambda: not self.mode_modify)
        r += 1
        FieldF (l,r,c, lab="Trailing edge", get=lambda: self.geo.te[0], width=75, dec=7, style=lambda: self._style (self.geo.te[0], 1.0))
        r += 1
        FieldF (l,r,c+1,get=lambda: self.geo.te[2], width=75, dec=7, style=lambda: self._style (self.geo.te[0], 1.0))

        r,c = 0, 2 
        SpaceC (l,c, width=10, stretch=0)
        c += 1 
        FieldF (l,r,c+1,get=lambda: self.geo.le[1], width=75, dec=7, style=lambda: self._style (self.geo.le[1], 0.0))
        r += 1
        FieldF (l,r,c+1,get=lambda: self.geo.le_real[1], width=75, dec=7, style=self._style_le_real,
                hide=lambda: not self.mode_modify)
        r += 1
        FieldF (l,r,c+1,get=lambda: self.geo.te[1], width=75, dec=7, style=lambda: self._style (self.geo.te[1], -self.geo.te[3]))
        r += 1
        FieldF (l,r,c+1,get=lambda: self.geo.te[3], width=75, dec=7, style=lambda: self._style (self.geo.te[3], -self.geo.te[1]))

        r += 1
        SpaceR (l,r, height=5)
        r += 1
        Label  (l,r,0,colSpan=5, get=self._messageText, style=style.COMMENT, height=(None,None))

        l.setColumnMinimumWidth (0,80)
        l.setColumnStretch (c+3,1)
        return l


    def _style_le_real (self):
        """ returns style.WARNING if LE spline isn't close to LE"""
        if self.geo.isLe_closeTo_le_real: 
            if self.geo.isBasic:
                return style.NORMAL
            else: 
                return style.NORMAL
        else: 
            return style.WARNING


    def _style (self, val, target_val):
        """ returns style.WARNING if val isn't target_val"""
        if val != target_val and not self.airfoil.isFlapped: 
            return style.WARNING
        else: 
            return style.NORMAL


    def _messageText (self): 

        text = []
        if not self.geo.isNormalized:
            if self.geo.isSplined and not self.geo.isLe_closeTo_le_real:
                text.append("- Leading edge of spline is not at 0,0")
            elif self.geo.le[0] != 0.0 or self.geo.le[1] != 1.0 : 
                text.append("- Leading edge is not at 0,0")
        if not self.airfoil.isFlapped:
            if self.geo.te[0] != 1.0 or self.geo.te[2] != 1.0 : 
                text.append("- Trailing edge is not at x=1.0")
            if self.geo.te[1] != -self.geo.te[3]: 
                text.append("- Trailing edge is not at y=0.0")

        if not text:
            if self.geo.isSymmetrical: 
                text.append("Airfoil is symmetrical")
            else: 
                text.append("Airfoil is normalized")

        text = '\n'.join(text)
        return text 



class Panel_Bezier (Panel_Airfoil_Abstract):
    """ Info about Bezier curves upper and lower  """

    name = 'Bezier'
    _width  = (180, None)


    @override
    @property
    def shouldBe_visible (self) -> bool:
        """ overloaded: only visible if geo is Bezier """
        return self.mode_bezier
    
    # ----

    @override
    @property
    def geo (self) -> Geometry_Bezier:
        return super().geo

    @property
    def upper (self) -> Side_Airfoil_Bezier:
        if self.geo.isBezier:
            return self.geo.upper

    @property
    def lower (self) -> Side_Airfoil_Bezier:
        if self.geo.isBezier:
            return self.geo.lower


    def _init_layout (self):

        l = QGridLayout()

        r,c = 0, 0 
        Label (l,r,c, get="Bezier control Points", colSpan=4)

        r += 1
        FieldI (l,r,c,   lab="Upper side", get=lambda: self.upper.nControlPoints,  width=50, step=1, lim=(3,10),
                         set=lambda n : self.geo.set_nControlPoints_of (self.upper, n))
        r += 1
        FieldI (l,r,c,   lab="Lower side",  get=lambda: self.lower.nControlPoints,  width=50, step=1, lim=(3,10),
                         set=lambda n : self.geo.set_nControlPoints_of (self.lower, n))

        r += 1
        SpaceR (l,r, height=10, stretch=2)
        l.setColumnMinimumWidth (0,70)
        l.setColumnStretch (c+2,4)
        
        return l
 


class Panel_Bezier_Match (Panel_Airfoil_Abstract):
    """ Match Bezier functions  """

    name = 'Bezier Match'
    _width  = (370, None)


    @override
    @property
    def shouldBe_visible (self) -> bool:
        """ overloaded: only visible if geo is Bezier """
        return self.mode_bezier

    # ----

    @property
    def upper (self) -> Side_Airfoil_Bezier:
        if self.geo.isBezier: return self.geo.upper

    @property
    def lower (self) -> Side_Airfoil_Bezier:
        if self.geo.isBezier: return self.geo.lower

    @property
    def curv_upper (self) -> Line:
        if self.geo.isBezier:
            return self.geo.curvature.upper

    @property
    def curv_lower (self) -> Line:
        if self.geo.isBezier:
            return self.geo.curvature.lower

    @property
    def curv (self) -> Curvature_Abstract:
        return self.geo.curvature

    @property
    def target_airfoil (self) -> Airfoil:
        return self.app.airfoil_org

    @property
    def target_upper (self) -> Line:
        if self.target_airfoil: return self.target_airfoil.geo.upper

    @property
    def target_lower (self) -> Line:
        if self.target_airfoil: return self.target_airfoil.geo.lower

    @property
    def target_curv_le (self) -> float:
        return self.target_airfoil.geo.curvature.best_around_le

    @property
    def max_curv_te_upper (self) -> Line:
        if self.target_airfoil: return self.target_airfoil.geo.curvature.at_upper_te

    @property
    def max_curv_te_lower (self) -> Line:
        if self.target_airfoil: return self.target_airfoil.geo.curvature.at_lower_te

    def norm2_upper (self): 
        """ norm2 deviation of airfoil to target - upper side """
        if self._norm2_upper is None: 
            self._norm2_upper = Line.norm2_deviation_to (self.upper.bezier, self.target_upper) 
        return  self._norm2_upper    


    def norm2_lower (self): 
        """ norm2 deviation of airfoil to target  - upper side """
        if self._norm2_lower is None: 
            self._norm2_lower = Line.norm2_deviation_to (self.lower.bezier, self.target_lower)  
        return self._norm2_lower


    def _init_layout (self):

        self._norm2_upper = None                                # cached value of norm2 deviation 
        self._norm2_lower = None                                # cached value of norm2 deviation 
        self._target_curv_le = None 
        self._target_curv_le_weighting = None

        l = QGridLayout()

        if self.target_airfoil is not None: 

            self._target_curv_le = self.target_airfoil.geo.curvature.best_around_le 

            r,c = 0, 0 
            Label  (l,r,c+1, get="Deviation", width=70)

            r += 1
            Label  (l,r,c,   get="Upper Side")
            FieldF (l,r,c+1, width=60, dec=3, unit="%", get=self.norm2_upper,
                             style=lambda: Match_Bezier_Dialog.style_deviation (self.norm2_upper()))
            r += 1
            Label  (l,r,c,   get="Lower Side")
            FieldF (l,r,c+1, width=60, dec=3, unit="%", get=self.norm2_lower,
                             style=lambda: Match_Bezier_Dialog.style_deviation (self.norm2_lower()))

            r,c = 0, 2 
            SpaceC(l,  c, width=5)
            c += 1
            Label (l,r,c, colSpan=2, get="LE curvature TE")
    
            r += 1
            FieldF (l,r,c  , get=lambda: self.curv_upper.max_xy[1], width=40, dec=0, 
                    style=lambda: Match_Bezier_Dialog.style_curv_le(self._target_curv_le, self.curv_upper))
            FieldF (l,r,c+1, get=lambda: self.curv_upper.te[1],     width=40, dec=1, 
                    style=lambda: Match_Bezier_Dialog.style_curv_te(self.max_curv_te_upper, self.curv_upper))

            r += 1
            FieldF (l,r,c  , get=lambda: self.curv_lower.max_xy[1], width=40, dec=0, 
                    style=lambda: Match_Bezier_Dialog.style_curv_le(self._target_curv_le, self.curv_lower))
            FieldF (l,r,c+1, get=lambda: self.curv_lower.te[1],     width=40, dec=1, 
                    style=lambda: Match_Bezier_Dialog.style_curv_te(self.max_curv_te_lower, self.curv_lower))

            r,c = 0, 5 
            SpaceC (l,  c, width=10)
            c += 1
            r += 1
            Button (l,r,c  , text="Match...", width=70,
                            set=lambda: self._match_bezier (self.upper, self.target_upper, 
                                                            self.target_curv_le, self.max_curv_te_upper))
            r += 1
            Button (l,r,c  , text="Match...", width=70,
                            set=lambda: self._match_bezier (self.lower, self.target_lower, 
                                                            self.target_curv_le, self.max_curv_te_lower))
            c = 0 
            r += 1
            SpaceR (l,r, height=5, stretch=2)
            r += 1
            Label  (l,r,0, get=self._messageText, colSpan=7, height=(40, None), style=style.COMMENT)
            l.setColumnMinimumWidth (0,70)
            l.setColumnStretch (c+6,2)

        else: 
            SpaceR (l,0)
            Label  (l,1,0, get="Select a target airfoil to match...", style=style.COMMENT)
            SpaceR (l,2, stretch=2)
        return l
 

    def _match_bezier (self, aSide : Side_Airfoil_Bezier, aTarget_line : Line, 
                            target_curv_le: float, max_curv_te : float  ): 
        """ run match bezier (dialog) """ 

        match_bezier = Match_Bezier_Dialog (self, aSide, aTarget_line,
                                    target_curv_le = target_curv_le,
                                    max_curv_te = max_curv_te,
                                    parentPos=(0.1, 0.1), dialogPos=(0,1))

        match_bezier.sig_new_bezier.connect     (self.app.sig_bezier_changed.emit)
        match_bezier.sig_pass_finished.connect  (self.app.sig_airfoil_changed.emit)
        match_bezier.sig_match_finished.connect (self._on_match_finished)

        # leave button press callback 
        timer = QTimer()                                
        timer.singleShot(10, lambda: match_bezier.exec())     # delayed emit 
       

    def _on_match_finished (self, aSide : Side_Airfoil_Bezier):
        """ slot for match Bezier finished - reset airfoil"""

        geo : Geometry_Bezier = self.geo
        geo.finished_change_of (aSide)              # will reset and handle changed  

        self.app._on_airfoil_changed()


    @override
    def refresh (self, reinit_layout=False):

        # reset cached deviations
        self._norm2_lower = None
        self._norm2_upper = None 
        super().refresh(reinit_layout=reinit_layout)
        

    def _messageText (self): 
        """ user warnings"""
        text = []
        r_upper_dev = Matcher.result_deviation (self.norm2_upper())
        r_lower_dev = Matcher.result_deviation (self.norm2_lower())

        r_upper_le = Matcher.result_curv_le (self._target_curv_le, self.curv_upper)
        r_lower_le = Matcher.result_curv_le (self._target_curv_le, self.curv_lower)
        r_upper_te = Matcher.result_curv_te (self.max_curv_te_upper,self.curv_upper)
        r_lower_te = Matcher.result_curv_te (self.max_curv_te_lower, self.curv_lower)

        is_bad = Matcher.result_quality.BAD
        if r_upper_dev == is_bad or r_lower_dev == is_bad:
           text.append("- Deviation is quite high")
        if r_upper_le == is_bad or r_lower_le == is_bad:
           text.append(f"- Curvature at LE differs too much from target ({int(self._target_curv_le)})")
        if r_upper_te == is_bad or r_lower_te == is_bad:
           text.append("- Curvature at TE is quite high")

        text = '\n'.join(text)
        return text 

